# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for the Dataset class.
"""

from typing import cast

import numpy as np
import pandas as pd
import scipy.sparse as sps
import torch
from numpy.typing import ArrayLike

from pytest import mark, raises

from lenskit.data import Dataset, FieldError, from_interactions_df
from lenskit.data.matrix import CSRStructure
from lenskit.testing import ml_ds, ml_ratings  # noqa: F401


def _check_user_offset_counts(ml_ds: Dataset, ml_ratings: pd.DataFrame, offsets: ArrayLike):
    user_counts = ml_ratings["user_id"].value_counts().reindex(ml_ds.users.index)
    row_lens = np.diff(offsets)
    assert np.all(row_lens == user_counts)


def _check_user_number_counts(ml_ds: Dataset, ml_ratings: pd.DataFrame, nums: ArrayLike):
    users, counts = np.unique(nums, return_counts=True)
    user_counts = ml_ratings["user_id"].value_counts().reindex(ml_ds.users.ids(users))
    assert np.all(counts == user_counts)


def _check_item_number_counts(ml_ds: Dataset, ml_ratings: pd.DataFrame, nums: ArrayLike):
    items, counts = np.unique(nums, return_counts=True)
    item_counts = ml_ratings["item_id"].value_counts().reindex(ml_ds.items.ids(items))
    assert np.all(counts == item_counts)


def _check_user_ids(ml_ds: Dataset, ml_ratings: pd.DataFrame, nums: ArrayLike):
    ml_ratings = ml_ratings.sort_values(["user_id", "item_id"])
    assert np.all(ml_ds.users.ids(np.asarray(nums)) == ml_ratings["user_id"])


def _check_item_ids(ml_ds: Dataset, ml_ratings: pd.DataFrame, nums: ArrayLike):
    ml_ratings = ml_ratings.sort_values(["user_id", "item_id"])
    assert np.all(ml_ds.items.ids(np.asarray(nums)) == ml_ratings["item_id"])


def _check_ratings(ml_ds: Dataset, ml_ratings: pd.DataFrame, rates: ArrayLike):
    ml_ratings = ml_ratings.sort_values(["user_id", "item_id"])
    assert np.all(rates == ml_ratings["rating"])


def _check_timestamp(ml_ds: Dataset, ml_ratings: pd.DataFrame, ts: ArrayLike):
    ml_ratings = ml_ratings.sort_values(["user_id", "item_id"])
    if ts.dtype.kind == "i":
        ts = pd.to_datetime(ts, unit="s")
    assert np.all(ts == ml_ratings["timestamp"])


def test_matrix_structure(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="structure")
    assert isinstance(log, CSRStructure)
    assert log.nnz == len(ml_ratings)

    assert log.nrows == ml_ratings["user_id"].nunique()
    assert log.ncols >= ml_ratings["item_id"].nunique()
    assert log.ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.rowptrs)
    _check_item_number_counts(ml_ds, ml_ratings, log.colinds)
    _check_item_ids(ml_ds, ml_ratings, log.colinds)


def test_matrix_structure_field_fail(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    with raises(ValueError, match="structure does not support fields"):
        ml_ds.interaction_matrix(format="structure", field="timestamp")  # type: ignore


def test_matrix_structure_coo_fail(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    with raises(ValueError, match="unsupported layout"):
        ml_ds.interaction_matrix(format="structure", layout="coo")  # type: ignore


def test_matrix_pandas(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="pandas", field="rating")
    assert isinstance(log, pd.DataFrame)
    assert len(log) == len(ml_ratings)
    assert "rating" in log.columns

    _check_user_number_counts(ml_ds, ml_ratings, log["user_num"])
    _check_user_ids(ml_ds, ml_ratings, log["user_num"])
    _check_item_number_counts(ml_ds, ml_ratings, log["item_num"])
    _check_item_ids(ml_ds, ml_ratings, log["item_num"])
    _check_ratings(ml_ds, ml_ratings, log["rating"])


def test_matrix_pandas_orig_id(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    "Test that Pandas can return original IDs."
    log = ml_ds.interaction_matrix(format="pandas", field="rating", original_ids=True)
    assert isinstance(log, pd.DataFrame)
    assert len(log) == len(ml_ratings)
    assert "user_id" in log.columns
    assert "item_id" in log.columns

    m2 = ml_ds.interaction_matrix(format="scipy", layout="coo")

    assert np.all(log["user_id"] == ml_ds.users.ids(m2.row))
    assert np.all(log["item_id"] == ml_ds.items.ids(m2.col))

    _check_ratings(ml_ds, ml_ratings, log["rating"])


def test_matrix_pandas_timestamp(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="pandas", field="timestamp")
    assert isinstance(log, pd.DataFrame)
    assert len(log) == len(ml_ratings)

    _check_user_number_counts(ml_ds, ml_ratings, log["user_num"])
    _check_user_ids(ml_ds, ml_ratings, log["user_num"])
    _check_item_number_counts(ml_ds, ml_ratings, log["item_num"])
    _check_item_ids(ml_ds, ml_ratings, log["item_num"])
    _check_timestamp(ml_ds, ml_ratings, log["timestamp"])


def test_matrix_pandas_unknown_field(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    with raises(FieldError, match=r"rating\[playcount\]"):
        ml_ds.interaction_matrix(format="pandas", field="playcount")  # type: ignore


def test_matrix_pandas_indicator(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="pandas")
    assert isinstance(log, pd.DataFrame)
    assert len(log) == len(ml_ratings)

    _check_user_number_counts(ml_ds, ml_ratings, log["user_num"])
    _check_user_ids(ml_ds, ml_ratings, log["user_num"])
    _check_item_number_counts(ml_ds, ml_ratings, log["item_num"])
    _check_item_ids(ml_ds, ml_ratings, log["item_num"])


@mark.parametrize("generation", ["modern", "legacy"])
def test_matrix_scipy_coo(ml_ratings: pd.DataFrame, ml_ds: Dataset, generation):
    log = ml_ds.interaction_matrix(
        format="scipy", layout="coo", field="rating", legacy=generation == "legacy"
    )
    assert isinstance(log, sps.coo_array if generation == "modern" else sps.coo_matrix)
    assert log.nnz == len(ml_ratings)

    nrows, ncols = cast(tuple[int, int], log.shape)
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    assert log.row.dtype == np.int32
    assert log.col.dtype == np.int32
    _check_user_number_counts(ml_ds, ml_ratings, log.row)
    _check_user_ids(ml_ds, ml_ratings, log.row)
    # ensure users are sorted
    assert np.all(np.diff(log.row) >= 0)

    _check_item_number_counts(ml_ds, ml_ratings, log.col)
    _check_item_ids(ml_ds, ml_ratings, log.col)
    _check_ratings(ml_ds, ml_ratings, log.data)


@mark.parametrize("generation", ["modern", "legacy"])
def test_matrix_scipy_csr(ml_ratings: pd.DataFrame, generation):
    ml_ds = from_interactions_df(ml_ratings.sample(frac=1.0, replace=False))
    log = ml_ds.interaction_matrix(format="scipy", field="rating", legacy=generation == "legacy")
    assert isinstance(log, sps.csr_array if generation == "modern" else sps.csr_matrix)
    assert log.nnz == len(ml_ratings)

    nrows, ncols = cast(tuple[int, int], log.shape)
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.indptr)
    _check_item_number_counts(ml_ds, ml_ratings, log.indices)
    _check_item_ids(ml_ds, ml_ratings, log.indices)
    _check_ratings(ml_ds, ml_ratings, log.data)


@mark.parametrize("generation", ["modern", "legacy"])
def test_matrix_scipy_timestamp(ml_ratings: pd.DataFrame, ml_ds: Dataset, generation):
    log = ml_ds.interaction_matrix(format="scipy", field="timestamp", legacy=generation == "legacy")
    assert isinstance(log, sps.csr_array if generation == "modern" else sps.csr_matrix)
    assert log.nnz == len(ml_ratings)

    nrows, ncols = cast(tuple[int, int], log.shape)
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.indptr)
    _check_item_number_counts(ml_ds, ml_ratings, log.indices)
    _check_item_ids(ml_ds, ml_ratings, log.indices)
    _check_timestamp(ml_ds, ml_ratings, log.data)


@mark.parametrize("generation", ["modern", "legacy"])
def test_matrix_scipy_indicator(ml_ratings: pd.DataFrame, ml_ds: Dataset, generation):
    log = ml_ds.interaction_matrix(format="scipy", legacy=generation == "legacy")
    assert isinstance(log, sps.csr_array if generation == "modern" else sps.csr_matrix)
    assert log.nnz == len(ml_ratings)

    nrows, ncols = cast(tuple[int, int], log.shape)
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.indptr)
    _check_item_number_counts(ml_ds, ml_ratings, log.indices)
    _check_item_ids(ml_ds, ml_ratings, log.indices)

    # right rating values
    assert np.all(log.data == 1.0)


def test_matrix_torch_csr(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="torch", field="rating")
    assert isinstance(log, torch.Tensor)
    assert log.is_sparse_csr
    assert log.values().shape == torch.Size([len(ml_ratings)])

    nrows, ncols = log.shape
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.crow_indices())
    _check_item_number_counts(ml_ds, ml_ratings, log.col_indices())
    _check_item_ids(ml_ds, ml_ratings, log.col_indices())
    _check_ratings(ml_ds, ml_ratings, log.values().numpy())

    assert log.crow_indices().dtype == torch.int64
    assert log.col_indices().dtype == torch.int32


def test_matrix_torch_indicator(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="torch")
    assert isinstance(log, torch.Tensor)
    assert log.is_sparse_csr
    assert log.values().shape == torch.Size([len(ml_ratings)])

    nrows, ncols = log.shape
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.crow_indices())
    _check_item_number_counts(ml_ds, ml_ratings, log.col_indices())
    _check_item_ids(ml_ds, ml_ratings, log.col_indices())
    assert np.all(log.values().numpy() == 1.0)


def test_matrix_torch_coo(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="torch", layout="coo", field="rating")
    assert isinstance(log, torch.Tensor)
    assert log.is_sparse
    assert log.values().shape == torch.Size([len(ml_ratings)])

    nrows, ncols = cast(tuple[int, int], log.shape)
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_number_counts(ml_ds, ml_ratings, log.indices()[0, :])
    _check_user_ids(ml_ds, ml_ratings, log.indices()[0, :])
    _check_item_number_counts(ml_ds, ml_ratings, log.indices()[1, :])
    _check_item_ids(ml_ds, ml_ratings, log.indices()[1, :])
    _check_ratings(ml_ds, ml_ratings, log.values().numpy())


def test_matrix_torch_timestamp(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    log = ml_ds.interaction_matrix(format="torch", field="timestamp")
    assert isinstance(log, torch.Tensor)
    assert log.is_sparse_csr
    assert log.values().shape == torch.Size([len(ml_ratings)])

    nrows, ncols = log.shape
    assert nrows == ml_ratings["user_id"].nunique()
    assert ncols >= ml_ratings["item_id"].nunique()
    assert ncols == ml_ds.item_count

    _check_user_offset_counts(ml_ds, ml_ratings, log.crow_indices())
    _check_item_number_counts(ml_ds, ml_ratings, log.col_indices())
    _check_item_ids(ml_ds, ml_ratings, log.col_indices())
    _check_timestamp(ml_ds, ml_ratings, log.values().numpy())


def test_matrix_rows_by_id(rng: np.random.Generator, ml_ratings: pd.DataFrame, ml_ds: Dataset):
    users = rng.choice(ml_ds.users.ids(), 50)

    for user in users:
        row = ml_ds.user_row(user)
        assert row is not None
        urows = ml_ratings[ml_ratings["user_id"] == user].sort_values("item_id")
        urows = urows.reset_index(drop=True)
        assert set(row.ids()) == set(urows["item_id"])
        assert np.all(row.numbers() == ml_ds.items.numbers(urows["item_id"]))

        ratings = row.field("rating")
        assert ratings is not None
        assert np.all(ratings == urows["rating"])

        timestamps = row.field("timestamp")
        assert timestamps is not None
        assert np.all(timestamps == urows["timestamp"])

        # we'll quick check additional fields on the item list here
        df = row.to_df()
        assert np.all(df["timestamp"] == urows["timestamp"])


def test_matrix_rows_by_num(rng: np.random.Generator, ml_ratings: pd.DataFrame, ml_ds: Dataset):
    users = rng.choice(ml_ds.user_count, 50)

    rated = set(zip(ml_ratings["user_id"], ml_ratings["item_id"]))
    rdf = ml_ds.interaction_matrix(format="pandas")
    rnums = set(zip(rdf["user_num"], rdf["item_num"]))

    dfi = ml_ratings.set_index(["user_id", "item_id"])

    for user in users:
        uid = ml_ds.users.id(user)
        row = ml_ds.user_row(user_num=user)
        assert row is not None
        assert row is not None
        urows = ml_ratings[ml_ratings["user_id"] == ml_ds.users.id(user)].sort_values("item_id")
        assert set(row.ids()) == set(urows["item_id"])

        assert np.all(row.numbers() == ml_ds.items.numbers(urows["item_id"]))
        assert all((user, ino) in rnums for ino in row.numbers())

        assert np.all(row.ids() == ml_ds.items.ids(row.numbers()))
        assert all((uid, item) in rated for item in row.ids())
        assert all((uid, item) in dfi.index for item in row.ids())

        ratings = row.field("rating")
        assert ratings is not None
        assert np.all(ratings == urows["rating"])

        timestamps = row.field("timestamp")
        assert timestamps is not None
        assert np.all(timestamps == urows["timestamp"])


def test_count_synthetic(ml_ds: Dataset):
    """
    Test that interaction matrices synthesize a count attribute.
    """

    matrix = ml_ds.interactions().matrix().scipy("count")
    assert isinstance(matrix, sps.csr_array)
    assert np.all(matrix.data == 1.0)


def test_matrix_csr_structure(ml_ds: Dataset):
    matrix = ml_ds.interactions().matrix()
    csr = matrix.csr_structure()
    assert csr.nrows == ml_ds.user_count
    assert csr.ncols == ml_ds.item_count
    assert csr.nnz == ml_ds.interaction_count


def test_matrix_csr_structure_arrow(ml_ds: Dataset):
    matrix = ml_ds.interactions().matrix()
    csr = matrix.csr_structure(format="arrow")
    assert len(csr) == matrix.n_rows
    assert csr.shape == (matrix.n_rows, matrix.n_cols)
    assert len(csr.indices) == matrix.count()
    assert not csr.has_values


def test_matrix_coo_structure(ml_ds: Dataset):
    matrix = ml_ds.interactions().matrix()
    coo = matrix.coo_structure()
    assert coo.nrows == ml_ds.user_count
    assert coo.ncols == ml_ds.item_count
    assert coo.nnz == ml_ds.interaction_count


def test_matrix_scipy_coo_partial(ml_ratings: pd.DataFrame, rng: np.random.Generator):
    ml_ratings = ml_ratings.copy()
    ml_ratings.loc[rng.choice(len(ml_ratings), 1000, replace=False), "rating"] = np.nan
    ds = from_interactions_df(ml_ratings)

    matrix = ds.interactions().matrix()

    coo = matrix.scipy(attribute="rating", layout="coo")
    assert coo.nnz == matrix.count() - 1000
    assert np.all(np.isfinite(coo.data))


def test_matrix_scipy_csr_partial(ml_ratings: pd.DataFrame, rng: np.random.Generator):
    ml_ratings = ml_ratings.copy()
    ml_ratings.loc[rng.choice(len(ml_ratings), 1000, replace=False), "rating"] = np.nan
    ds = from_interactions_df(ml_ratings)

    matrix = ds.interactions().matrix()

    coo = matrix.scipy(attribute="rating", layout="coo")
    assert coo.nnz == matrix.count() - 1000
    assert np.all(np.isfinite(coo.data))


def test_matrix_tensor_partial(ml_ratings: pd.DataFrame, rng: np.random.Generator):
    ml_ratings = ml_ratings.copy()
    ml_ratings.loc[rng.choice(len(ml_ratings), 1000, replace=False), "rating"] = np.nan
    ds = from_interactions_df(ml_ratings)

    matrix = ds.interactions().matrix()

    coo = matrix.torch(attribute="rating")
    assert len(coo.values()) == matrix.count() - 1000
    assert torch.all(torch.isfinite(coo.values()))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from itertools import product

import numpy as np

from pytest import mark

from lenskit.data import Dataset
from lenskit.logging import get_logger

_log = get_logger(__name__)


@mark.parametrize("weighting", ["uniform", "popularity"])
def test_negative(rng: np.random.Generator, ml_ds: Dataset, weighting):
    log = _log.bind()
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 100, replace=True)
    users = np.require(users, "i4")

    negatives = matrix.sample_negatives(users, rng=rng, weighting=weighting)
    assert negatives.shape == (100,)

    log.info("checking basic item results")
    assert np.all(negatives >= 0)
    assert np.all(negatives < ml_ds.item_count)
    log.info("checking negative items")
    for u, i in zip(users, negatives):
        ulog = log.bind(
            user_num=u.item(),
            user_id=int(ml_ds.users.id(u)),  # type: ignore
            item_num=i.item(),
            item_id=int(ml_ds.items.id(i)),  # type: ignore
        )
        row = ml_ds.user_row(user_num=u)
        ulog = ulog.bind(u_nitems=len(row))
        ulog.debug("checking if item is negative")
        assert not matrix._coords.contains(u, i)
        print(ml_ds.users.id(u), row.ids())
        assert i not in row.numbers()


@mark.parametrize("weighting", ["uniform", "popularity"])
def test_negative_multiple(rng: np.random.Generator, ml_ds: Dataset, weighting):
    log = _log.bind()
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 100, replace=True)
    users = np.require(users, "i4")

    negatives = matrix.sample_negatives(users, rng=rng, weighting=weighting, n=2)
    assert negatives.shape == (100, 2)

    log.info("checking basic item results")
    assert np.all(negatives >= 0)
    assert np.all(negatives < ml_ds.item_count)
    log.info("checking negative items")
    for c in range(2):
        for u, i in zip(users, negatives[:, c]):
            ulog = log.bind(
                user_num=u.item(),
                user_id=int(ml_ds.users.id(u)),  # type: ignore
                item_num=i.item(),
                item_id=int(ml_ds.items.id(i)),  # type: ignore
            )
            row = ml_ds.user_row(user_num=u)
            ulog = ulog.bind(u_nitems=len(row))
            ulog.debug("checking if item is negative")
            assert not matrix._coords.contains(u, i)
            print(ml_ds.users.id(u), row.ids())
            assert i not in row.numbers()


@mark.parametrize("weighting", ["uniform", "popularity"])
def test_negative_unverified(rng: np.random.Generator, ml_ds: Dataset, weighting):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 500, replace=True)
    users = np.require(users, "i4")

    negatives = matrix.sample_negatives(users, verify=False, rng=rng, weighting=weighting)

    assert np.all(negatives >= 0)
    assert np.all(negatives < ml_ds.item_count)


@mark.benchmark()
def test_negative_unverified_bench(rng: np.random.Generator, ml_ds: Dataset, benchmark):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 500, replace=True)
    users = np.require(users, "i4")

    def sample():
        _items = matrix.sample_negatives(users, verify=False, rng=rng)

    benchmark(sample)


@mark.benchmark()
@mark.parametrize("count", [500, 5000, 8192])
def test_negative_verified_bench(count, rng: np.random.Generator, ml_ds: Dataset, benchmark):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, count, replace=True)
    users = np.require(users, "i4")

    def sample():
        _items = matrix.sample_negatives(users, rng=rng)

    benchmark(sample)


@mark.benchmark()
def test_negative_multiple_bench(rng: np.random.Generator, ml_ds: Dataset, benchmark):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 500, replace=True)
    users = np.require(users, "i4")

    def sample():
        _items = matrix.sample_negatives(users, n=10, rng=rng)

    benchmark(sample)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import os
import warnings

import structlog
import torch
from numpy.random import Generator, default_rng

from hypothesis import settings
from pytest import fixture, register_assert_rewrite, skip

from lenskit.parallel import ensure_parallel_init
from lenskit.random import init_global_rng

register_assert_rewrite("lenskit.testing")

# bring common fixtures into scope
from lenskit.testing import ml_100k, ml_ds, ml_ds_unchecked, ml_ratings  # noqa: E402, F401

logging.getLogger("numba").setLevel(logging.INFO)

_log = structlog.stdlib.get_logger("lenskit.tests")
RNG_SEED = 42
if "LK_TEST_FREE_RNG" in os.environ:
    warnings.warn("using nondeterministic RNG initialization")
    RNG_SEED = None

structlog.configure(
    [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.MaybeTimeStamper(fmt="iso"),
        structlog.processors.KeyValueRenderer(key_order=["timestamp", "event"]),
    ],
    wrapper_class=structlog.stdlib.BoundLogger,
    logger_factory=structlog.stdlib.LoggerFactory(),
    cache_logger_on_first_use=True,
)


@fixture
def rng() -> Generator:
    if RNG_SEED is None:
        return default_rng()
    else:
        return default_rng(RNG_SEED)


@fixture(autouse=True)
def init_rng(request):
    if RNG_SEED is not None:
        init_global_rng(RNG_SEED)


@fixture(scope="module", params=["cpu", "cuda"])
def torch_device(request):
    """
    Fixture for testing across Torch devices.  This fixture is parameterized, so
    if you write a test function with a parameter ``torch_device`` as its first
    parameter, it will be called once for each available Torch device.
    """
    dev = request.param
    if dev == "cuda" and not torch.cuda.is_available():
        skip("CUDA not available")
    if dev == "mps" and not torch.backends.mps.is_available():
        skip("MPS not available")
    yield dev


@fixture(autouse=True)
def log_test(request):
    try:
        modname = request.module.__name__ if request.module else "<unknown>"
    except Exception:
        modname = "<unknown>"
    funcname = request.function.__name__ if request.function else "<unknown>"
    _log.info("running test %s:%s", modname, funcname)


def pytest_collection_modifyitems(items):
    # add 'slow' to all 'eval' tests
    for item in items:
        evm = item.get_closest_marker("eval")
        rdm = item.get_closest_marker("realdata")
        slm = item.get_closest_marker("slow")
        if slm is None and (evm is not None or rdm is not None):
            _log.debug("adding slow mark to %s", item)
            item.add_marker("slow")


settings.register_profile("default", deadline=1000)
ensure_parallel_init()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pytest import approx

from lenskit.flexmf import FlexMFExplicitScorer
from lenskit.flexmf._explicit import FlexMFExplicitConfig
from lenskit.testing import BasicComponentTests, ScorerTests


class TestFlexMFExplicitL2(BasicComponentTests, ScorerTests):
    expected_rmse = approx(0.96, abs=0.05)
    component = FlexMFExplicitScorer
    config = FlexMFExplicitConfig(reg_method="L2")


class TestFlexMFExplicitAdam(BasicComponentTests, ScorerTests):
    component = FlexMFExplicitScorer
    config = FlexMFExplicitConfig(reg_method="AdamW")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pickle
import sys
from pathlib import Path

import click
from xopen import xopen

import lenskit.operations as ops
from lenskit import batch
from lenskit.data import Dataset, ItemList, ListILC, UserIDKey
from lenskit.logging import Stopwatch, get_logger, item_progress
from lenskit.pipeline import PipelineProfiler
from lenskit.random import random_generator

_log = get_logger(__name__)


@click.command("recommend")
@click.option(
    "-o",
    "--output",
    "out_file",
    metavar="FILE",
    help="Output file for recommendations.",
)
@click.option("--print/--no-print", "print_recs", default=True, help="Print recommendations.")
@click.option("-n", "--list-length", type=int, help="Recommendation list length.")
@click.option("--batch/--no-batch", "use_batch", default=False, help="Use batch.recommend.")
@click.option("-j", "--process-count", type=int, help="Use specified number of worker processes.")
@click.option("--ray", "use_ray", is_flag=True, help="Ue Ray for parallelism.")
@click.option("-d", "--dataset", metavar="DATA", type=Path, help="Use dataset DATA.")
@click.option("-u", "--users-file", type=Path, metavar="FILE", help="Load list of users from FILE.")
@click.option("--random-users", type=int, metavar="N", help="Recommend for N random users.")
@click.option("--profile", type=Path, metavar="FILE", help="Profile profile inference.")
@click.argument("PIPE_FILE", type=Path)
@click.argument("USERS", nargs=-1)
def recommend(
    out_file: Path,
    users_file: Path | None,
    print_recs: bool,
    use_batch: bool,
    process_count: int | None,
    use_ray: bool,
    random_users: int | None,
    list_length: int | None,
    dataset: Path | None,
    profile: Path | None,
    pipe_file: Path,
    users: list,
):
    """
    Generate recommendations from a serialized recommendation pipeline.
    """
    _log.warning("the recommend CLI is experimental and may change without notice")

    _log.info("loading pipeline", file=str(pipe_file))
    with xopen(pipe_file, "rb", threads=0) as pf:
        pipe = pickle.load(pf)
    log = _log.bind(name=pipe.name)

    if dataset is not None:
        data = Dataset.load(dataset)
        log = log.bind(data=data.name)
    else:
        data = None

    if random_users is not None:
        if data is None:
            log.error("dataset required for random users")
            sys.exit(5)
        rng = random_generator()
        log.info("selecting random users", count=random_users)
        users = rng.choice(data.users.ids(), random_users)  # type: ignore

    profiler = None
    if profile:
        profiler = PipelineProfiler(pipe, profile)

    if use_batch or use_ray or process_count is not None:
        if use_ray:
            n_jobs = "ray"
        else:
            n_jobs = process_count

        all_recs = batch.recommend(pipe, users, list_length, n_jobs=n_jobs, profiler=profiler)
    else:
        timer = Stopwatch(start=False)
        all_recs = None if out_file is None else ListILC(UserIDKey)
        with item_progress("user recommendations", len(users)) as pb:
            for user in users:
                ulog = log.bind(user=user)
                ulog.debug("generating single-user recommendations")
                with timer.measure(accumulate=True):
                    recs = ops.recommend(pipe, user, list_length, profiler=profiler)
                ulog.info(
                    "recommended for user",
                    length=len(recs),
                    time="{:.1f}ms".format(timer.elapsed(accumulated=False) * 1000),
                )

                if all_recs is not None:
                    all_recs.add(recs, user_id=user)

                if print_recs:
                    print_recommendation_list(recs, data)

                pb.update()

        log.info(
            "finished recommending for %d users in %s (%.1fms/u)",
            len(users),
            timer,
            timer.elapsed() * 1000 / len(users),
        )

    if profiler is not None:
        profiler.close()

    if out_file is not None:
        assert all_recs is not None
        log.info("saving recommendations to %s", str(out_file), count=len(all_recs))
        all_recs.save_parquet(out_file)


def print_recommendation_list(recs: ItemList, data: Dataset | None):
    titles = None
    if data is not None:
        items = data.entities("item")
        if "title" in items.attributes:
            titles = items.select(ids=recs.ids()).attribute("title").pandas()

    for item in recs.ids():
        if titles is not None:
            print("item {}: {}".format(item, titles.loc[item]))
        else:
            print("item {}".format(item))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Sequence

from lenskit.data import GenericKey, ItemListCollection


class BatchResults:
    """
    Results from a batch recommendation run.  Results consist of the outputs of
    various pipeline components for each of the test users.  Results may be
    ``None``, if the pipeline produced no output for that user.

    Stability:
        Caller
    """

    _key_schema: type[tuple] | Sequence[str]
    _data: dict[str, ItemListCollection[GenericKey]]

    def __init__(self, key: type[tuple] | Sequence[str]):
        """
        Construct a new set of batch results.
        """
        self._key_schema = key
        self._data = {}

    @property
    def outputs(self) -> list[str]:
        """
        Get the list of output names in these results.
        """
        return list(self._data.keys())

    def output(self, name: str) -> ItemListCollection[GenericKey]:
        """
        Get the item lists for a particular output component.

        Args:
            name:
                The output name. This may or may not be the same as the
                component name.
        """
        return self._data[name]

    def add_result(self, name: str, key: GenericKey, result: object):
        """
        Add a single result for one of the outputs.

        Args:
            name:
                The output name in which to save this result.
            user:
                The user identifier for this result.
            result:
                The result object to save.
        """

        if name not in self._data:
            self._data[name] = ItemListCollection(self._key_schema)

        try:
            self._data[name].add(result, *key)
        except TypeError as e:
            raise TypeError(f"invalid key {key} (for type {self._data[name].key_type})", e)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import warnings

from lenskit.logging import Stopwatch
from lenskit.random import derivable_rng, random_generator, set_global_rng

warnings.warn("lenskit.util is deprecated, import from original modules", DeprecationWarning)

__all__ = [
    "Stopwatch",
    "derivable_rng",
    "random_generator",
    "set_global_rng",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import math
import platform
from abc import abstractmethod
from collections.abc import Callable, Generator, Sequence
from dataclasses import dataclass, field, replace
from typing import Mapping

import numpy as np
import structlog
import torch
from torch import Tensor
from typing_extensions import TYPE_CHECKING, Generic, Self, TypeVar

from lenskit.data import Dataset, MatrixRelationshipSet
from lenskit.logging import get_logger, item_progress
from lenskit.parallel.config import ensure_parallel_init
from lenskit.training import ModelTrainer, TrainingOptions

from ._model import FlexMFModel

try:
    from torch._dynamo.exc import TorchDynamoException
except ImportError:
    TorchDynamoException = None

# hide base import to avoid circular imports
if TYPE_CHECKING:
    from ._base import FlexMFConfigBase, FlexMFScorerBase


Comp = TypeVar("Comp", bound="FlexMFScorerBase", default="FlexMFScorerBase")
Cfg = TypeVar("Cfg", bound="FlexMFConfigBase", default="FlexMFConfigBase")

_log = get_logger(__name__)


class FlexMFTrainerBase(ModelTrainer, Generic[Comp, Cfg]):
    """
    Trainer for a FlexMF model.  This class should be inherited by the
    model-specific trainers.
    """

    component: Comp
    """
    The component whose model is being trained.
    """
    opt: torch.optim.Optimizer
    """
    The PyTorch optimizer.
    """
    data: FlexMFTrainingData
    """
    The training data, set up for the training process.
    """
    epochs_trained: int = 0
    """
    The number of epochs trained so far.
    """

    device: str
    """
    Torch device for training.
    """

    explicit_norm: bool
    """
    Use explicit (L2) norm for normalizing data.
    """

    rng: np.random.Generator
    """
    NumPy generator for random number generation.
    """
    torch_rng: torch.Generator
    """
    PyPTorch generator for random number generation.
    """

    log: structlog.stdlib.BoundLogger = field(default_factory=lambda: _log.bind())
    """
    A logger, that is bound the current training status / position.
    """

    _compiled_model: Callable[..., torch.Tensor] | None = None

    def __init__(self, component: Comp, data: Dataset, options: TrainingOptions):
        ensure_parallel_init()

        self.component = component
        self.explicit_norm = component.config.reg_method == "L2"

        self.log = _log.bind(scorer=self.__class__.__name__, size=self.config.embedding_size)

        self.device = options.configured_device(gpu_default=True)
        self.rng = options.random_generator()
        self.torch_rng = options.random_generator(type="torch")

        self.data = self.prepare_data(data)
        self.component.model = self.create_model()

        # zero out non-interacted users/items
        users = data.user_stats()
        self.model.zero_users(torch.tensor(users["count"].values == 0))
        items = data.item_stats()
        self.model.zero_items(torch.tensor(items["count"].values == 0))

        self.log.info("preparing to train %r", self.component, device=self.device)
        self.component.model = self.model.to(self.device)
        self.model.train(True)

        if platform.system() not in ("Linux", "Darwin"):
            _log.warn("compiled models are only usable on Linux and macOS")
        elif torch.__version__ < "2.8":
            _log.warn("compiled models require Torch >=2.8")
        else:
            _log.debug("compiling FlexMF model")
            self._compiled_model = torch.compile(self.model)

        self.setup_optimizer()

    @property
    def config(self) -> Cfg:
        """
        Get the component configuration.
        """
        return self.component.config  # type: ignore

    @property
    def model(self) -> FlexMFModel:
        """
        Get model being trained.
        """
        return self.component.model

    def call_model(self, *args, **kwargs) -> torch.Tensor:
        """
        Invoke the model, using the compiled version if available.
        """
        if self._compiled_model is not None:
            assert TorchDynamoException is not None
            try:
                return self._compiled_model(*args, **kwargs)
            except TorchDynamoException as e:
                self.log.warning("calling compiled model failed, falling back: %s", e)
                self._compiled_model = None
                return self.call_model(*args, **kwargs)
        else:
            return self.model(*args, **kwargs)

    def train_epoch(self) -> dict[str, float]:
        epoch = self.epochs_trained + 1
        self.log = elog = self.log.bind(epoch=epoch)
        elog.debug("creating epoch training data")
        epoch_data = self.epoch_data()

        tot_loss = torch.tensor(0.0).to(self.device)
        avg_loss = np.nan
        with item_progress(
            f"Training epoch {epoch}", epoch_data.batch_count, {"loss": ".3f"}
        ) as pb:
            elog.debug("beginning epoch")
            for i, batch in enumerate(epoch_data.batches(), 1):
                self.log = blog = elog.bind(batch=i)
                blog.debug("training batch")
                loss = self.train_batch(batch)
                self.opt.zero_grad()

                if i % 20 == 0:
                    avg_loss = tot_loss.item() / epoch_data.batch_count
                pb.update(loss=avg_loss)
                tot_loss += loss

        avg_loss = tot_loss.item() / epoch_data.batch_count
        elog.debug("epoch complete", loss=avg_loss)
        self.epochs_trained += 1
        return {"loss": avg_loss}

    @abstractmethod
    def train_batch(self, batch: FlexMFTrainingBatch) -> float:  # pragma: nocover
        """
        Compute and apply updates for a single batch.

        Args:
            batch:
                The training minibatch.

        Returns:
            The loss.
        """

    def finalize(self):
        del self.opt
        del self.data
        self.model.eval()

    def epoch_data(self) -> FlexMFTrainingEpoch:
        """
        Get training data for a single training epoch.
        """
        # permute the data
        perm = np.require(self.rng.permutation(self.data.n_samples), dtype=np.int32)
        # convert to tensor, send to the training data's device.
        if isinstance(self.data.items, torch.Tensor):
            perm = torch.tensor(perm).to(self.data.items.device)

        return FlexMFTrainingEpoch(self.data, perm)

    @abstractmethod
    def prepare_data(self, data: Dataset) -> FlexMFTrainingData:  # pragma: nocover
        """
        Set up the training data and context for the scorer.

        This method _returns_ the data object; the :meth:`setup` method will
        save its return value in :attr:`data`.
        """
        raise NotImplementedError()

    @abstractmethod
    def create_model(self) -> FlexMFModel:  # pragma: nocover
        """
        Create and initialize the model for training.

        This method should _return_ the model; the :meth:`setup` method will put
        the model in the component and arrange for it to be available via
        :attr:`model`.
        """
        raise NotImplementedError()

    def setup_optimizer(self):
        """
        Create the appropriate optimizer depending on the regularization method.
        """
        if self.config.reg_method == "AdamW":
            self.log.debug("creating AdamW optimizer")
            opt = torch.optim.AdamW(
                self.model.parameters(),
                lr=self.config.learning_rate,
                weight_decay=self.config.regularization,
            )
        else:
            self.log.debug("creating SparseAdam optimizer")
            opt = torch.optim.SparseAdam(self.model.parameters(), lr=self.config.learning_rate)
        self.opt = opt

    def get_parameters(self) -> Mapping[str, object]:
        return self.model.state_dict()

    def load_parameters(self, state: Mapping[str, object]) -> None:
        self.model.load_state_dict(state)


@dataclass
class FlexMFTrainingData:
    """
    Training data for a FlexMF model.

    Stability:
        Experimental
    """

    batch_size: int

    n_users: int
    n_items: int

    users: torch.Tensor | np.ndarray[tuple[int], np.dtype[np.int32]]
    "User numbers for training samples."
    items: torch.Tensor | np.ndarray[tuple[int], np.dtype[np.int32]]
    "Item numbers for training samples."

    matrix: MatrixRelationshipSet | None = None
    """
    The original relationship set we are training on.
    """

    fields: dict[str, torch.Tensor] = field(default_factory=dict)
    "Additional per-sample data fields."

    def to(self, device: str) -> Self:
        """
        Move this data to another device.
        """
        ut = torch.as_tensor(self.users, device=device)
        it = torch.as_tensor(self.items, device=device)
        fts = {f: torch.as_tensor(t, device=device) for (f, t) in self.fields.items()}
        return replace(self, users=ut, items=it, fields=fts)

    @property
    def n_samples(self) -> int:
        return len(self.users)


@dataclass
class FlexMFTrainingEpoch:
    """
    Permuted data for a single training epoch.
    """

    data: FlexMFTrainingData
    permutation: torch.Tensor | np.ndarray[tuple[int], np.dtype[np.int32]]

    @property
    def n_samples(self) -> int:
        return self.data.n_samples

    @property
    def batch_size(self) -> int:
        return self.data.batch_size

    @property
    def batch_count(self) -> int:
        return math.ceil(self.n_samples / self.batch_size)

    def batches(
        self, fields: Sequence[str] | set[str] | None = None
    ) -> Generator[FlexMFTrainingBatch, None, None]:
        """
        Iterate over batches in this epoch.

        Args:
            fields:
                The fields to include in the batch.
        """

        fields = set(self.data.fields.keys()) if fields is None else fields
        for start in range(0, self.n_samples, self.batch_size):
            end = min(start + self.batch_size, self.n_samples)
            rows = self.permutation[start:end]

            yield self.make_batch(rows, fields)

    def make_batch(
        self,
        rows: Tensor | np.ndarray[tuple[int], np.dtype[np.int32]],
        fields: Sequence[str] | set[str],
    ) -> FlexMFTrainingBatch:
        ut = self.data.users[rows]
        it = self.data.items[rows]
        fts = {f: self.data.fields[f][rows] for f in fields}
        return FlexMFTrainingBatch(self.data, ut, it, fts)


@dataclass
class FlexMFTrainingBatch:
    "Representation of a single batch."

    data: FlexMFTrainingData
    users: torch.Tensor | np.ndarray[tuple[int], np.dtype[np.int32]]
    items: torch.Tensor | np.ndarray[tuple[int], np.dtype[np.int32]]
    fields: dict[str, torch.Tensor]

    def to(self, device: str) -> Self:
        """
        Move this data to another device.
        """
        ut = torch.as_tensor(self.users, device=device)
        it = torch.as_tensor(self.items, device=device)
        fts = {f: torch.as_tensor(t, device=device) for (f, t) in self.fields.items()}
        return replace(self, users=ut, items=it, fields=fts)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from __future__ import annotations

from collections.abc import Iterable
from dataclasses import replace
from graphlib import TopologicalSorter
from os import PathLike
from pathlib import Path
from uuid import NAMESPACE_URL, uuid5

from numpy.random import BitGenerator, Generator, SeedSequence
from typing_extensions import (
    TYPE_CHECKING,
    Any,
    Literal,
    Mapping,
    TypeAlias,
    TypeVar,
    overload,
)

from lenskit.data import Dataset
from lenskit.diagnostics import PipelineError
from lenskit.logging import get_logger

from . import config
from ._hooks import RunHooks, default_run_hooks
from .config import PipelineConfig
from .nodes import (
    ComponentConstructorNode,
    ComponentInstanceNode,
    Node,
)
from .state import PipelineState

if TYPE_CHECKING:
    from lenskit.training import TrainingOptions

    from ._profiling import ProfileSink
    from .builder import PipelineBuilder

_log = get_logger(__name__)

# common type var for quick use
T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")

NAMESPACE_LITERAL_DATA = uuid5(NAMESPACE_URL, "https://ns.lenskit.org/literal-data/")
CloneMethod: TypeAlias = Literal["config", "pipeline-config"]


class Pipeline:
    """
    LensKit recommendation pipeline.  This is the core abstraction for using
    LensKit models and other components to produce recommendations in a useful
    way.  It allows you to wire together components in (mostly) abitrary graphs,
    train them on data, and serialize pipelines to disk for use elsewhere.

    Pipelines should not be directly instantiated; they must be built with a
    :class:`~lenskit.pipeline.PipelineBuilder` class, or loaded from a
    configuration with :meth:`from_config`. If you have a scoring model and just
    want to generate recommenations with a default setup and minimal
    configuration, see :func:`~lenskit.pipeline.topn_pipeline` or
    :class:`~lenskit.pipeline.RecPipelineBuilder`.

    Pipelines are also :class:`~lenskit.training.Trainable`, and train all
    trainable components.

    Stability:
        Caller
    """

    _config: config.PipelineConfig
    _nodes: dict[str, Node[Any]]
    _edges: dict[str, dict[str, str]]
    _aliases: dict[str, Node[Any]]
    _default: Node[Any] | None = None
    _hash: str | None = None
    _run_hooks: RunHooks

    def __init__(
        self,
        config: config.PipelineConfig,
        nodes: Iterable[Node[Any]],
        *,
        run_hooks: RunHooks,
    ):
        self._nodes = {}
        for node in nodes:
            if isinstance(node, ComponentConstructorNode):
                raise RuntimeError("pipeline is not fully instantiated")
            self._nodes[node.name] = node
        self._edges = {name: cc.inputs for (name, cc) in config.components.items()}

        self._config = config
        self._aliases = {}
        for a, t in config.aliases.items():
            self._aliases[a] = self.node(t)
        if config.default:
            self._default = self.node(config.default)

        self._run_hooks = default_run_hooks()
        for name, hooks in run_hooks.items():
            my_hooks = self._run_hooks.setdefault(name, [])  # type: ignore
            my_hooks.extend(hooks)  # type: ignore
            my_hooks.sort(key=lambda h: h.priority)

    @property
    def config(self) -> PipelineConfig:
        """
        Get the pipline configuration.

        .. important::

            Do not modify the configuration returned, or it will become
            out-of-sync with the pipeline and likely not behave correctly.
        """
        return self._config

    @property
    def name(self) -> str | None:
        """
        Get the pipeline name (if configured).
        """
        return self._config.meta.name

    @property
    def version(self) -> str | None:
        """
        Get the pipeline version (if configured).
        """
        return self._config.meta.version

    def meta(self) -> config.PipelineMeta:
        """
        Get the metadata (name, version, hash, etc.) for this pipeline without
        returning the whole config.
        """
        return self._config.meta

    def nodes(self) -> list[Node[object]]:
        """
        Get the nodes in the pipeline graph.
        """
        return list(self._nodes.values())

    @overload
    def node(self, node: str, *, missing: Literal["error"] = "error") -> Node[object]: ...
    @overload
    def node(self, node: str, *, missing: Literal["none"] | None) -> Node[object] | None: ...
    @overload
    def node(self, node: Node[T]) -> Node[T]: ...
    def node(
        self, node: str | Node[Any], *, missing: Literal["error", "none"] | None = "error"
    ) -> Node[object] | None:
        """
        Get the pipeline node with the specified name.  If passed a node, it
        returns the node or fails if the node is not a member of the pipeline.

        Args:
            node:
                The name of the pipeline node to look up, or a node to check for
                membership.

        Returns:
            The pipeline node, if it exists.

        Raises:
            KeyError:
                The specified node does not exist.
        """
        if isinstance(node, Node):
            self._check_member_node(node)
            return node
        elif node in self._aliases:
            return self._aliases[node]
        elif node in self._nodes:
            return self._nodes[node]
        elif missing == "none" or missing is None:
            return None
        else:
            raise KeyError(node)

    def component_names(self) -> list[str]:
        """
        Get the component names (in topological order).
        """
        sort = TopologicalSorter[str]()
        comp_nodes = [n for n in self.nodes() if isinstance(n, ComponentInstanceNode)]
        for node in comp_nodes:
            edges = self._edges.get(node.name, {})
            prev = [v for v in edges.values() if isinstance(self.node(v), ComponentInstanceNode)]
            sort.add(node.name, *prev)

        return list(sort.static_order())

    def node_input_connections(self, node: str | Node[Any]) -> Mapping[str, Node[Any]]:
        """
        Get the input wirings for a node.
        """
        node = self.node(node)
        edges = self._edges.get(node.name, {})
        return {name: self.node(src) for (name, src) in edges.items()}

    def component(self, node: str | Node[Any]) -> object | None:
        """
        Get the component at a particular node, if any.
        """
        node = self.node(node)
        if isinstance(node, ComponentInstanceNode):
            return node.component
        else:
            return None

    def clone(self) -> Pipeline:
        """
        Clone the pipeline, **without** its trained parameters.

        Returns:
            A new pipeline with the same components and wiring, but fresh
            instances created by round-tripping the configuration.
        """
        return self.from_config(self._config)

    @property
    def config_hash(self) -> str:
        """
        Get a hash of the pipeline's configuration to uniquely identify it for
        logging, version control, or other purposes.

        The hash format and algorithm are not guaranteed, but is stable within a
        LensKit version.  For the same version of LensKit and component code,
        the same configuration will produce the same hash, so long as there are
        no literal nodes.  Literal nodes will *usually* hash consistently, but
        since literals other than basic JSON values are hashed by pickling, hash
        stability depends on the stability of the pickle bytestream.

        In LensKit 2025.1, the configuration hash is computed by computing the
        JSON serialization of the pipeline configuration *without* a hash and
        returning the hex-encoded SHA256 hash of that configuration.
        """
        assert self._config.meta.hash, "pipeline configuration has no hash"
        return self._config.meta.hash

    @staticmethod
    def from_config(config: object, file_path: Path | None = None) -> Pipeline:
        """
        Reconstruct a pipeline from a serialized configuration.

        Args:
            config:
                The configuration object, as loaded from JSON, TOML, YAML, or
                similar. Will be validated into a :class:`PipelineConfig`.
            file_path:
                The path of the file from which the configuration was read,
                when available.

        Returns:
            The configured (but not trained) pipeline.

        Raises:
            PipelineError:
                If there is a configuration error reconstructing the pipeline.

        Warns:
            PipelineWarning:
                If the configuration is funny but usable; for example, the
                configuration includes a hash but the constructed pipeline does
                not have a matching hash.
        """
        from .builder import PipelineBuilder

        config = PipelineConfig.model_validate(config)
        builder = PipelineBuilder.from_config(config, file_path=file_path)
        return builder.build()

    @classmethod
    def load_config(cls, cfg_file: Path | PathLike[str]) -> Pipeline:
        """
        Load a pipeline from a saved configuration file.

        Args:
            cfg_file:
                The path to a TOML, YAML, or JSON file containing the pipeline
                configuration.

        Returns:
            The consructed pipeline.

        See Also:
            :meth:`from_config` for the actual pipeline instantiation logic.
        """
        from .builder import PipelineBuilder

        bld = PipelineBuilder.load_config(cfg_file)
        return bld.build()

    def modify(self) -> PipelineBuilder:
        """
        Create a pipeline builder from this pipeline in order to modify it.

        Pipelines cannot be modified in-place, but this method sets up a new
        builder that will create a modified copy of the pipeline.  Unmodified
        component instances are reused as-is.

        .. note::

            Since default connections are applied in
            :meth:`~lenskit.pipeline.PipelineBuilder.build`, the modifying
            builder does not have default connections.
        """
        from .builder import PipelineBuilder

        return PipelineBuilder.from_pipeline(self)

    def train(self, data: Dataset, options: TrainingOptions | None = None) -> None:
        """
        Trains the pipeline's trainable components (those implementing the
        :class:`TrainableComponent` interface) on some training data.

        .. admonition:: Random Number Generation
            :class: note

            If :attr:`TrainingOptions.rng` is set and is not a generator or bit
            generator (i.e. it is a seed), then this method wraps the seed in a
            :class:`~numpy.random.SeedSequence` and calls
            :class:`~numpy.random.SeedSequence.spawn()` to generate a distinct
            seed for each component in the pipeline.

        Args:
            data:
                The dataset to train on.
            options:
                The training options.  If ``None``, default options are used.
        """
        from lenskit.training import Trainable, TrainingOptions  # lazy import

        log = _log.bind(pipeline=self.name)
        if options is None:
            options = TrainingOptions()

        if isinstance(options.rng, SeedSequence):
            seed = options.rng
        elif options.rng is None or isinstance(options.rng, (Generator, BitGenerator)):
            seed = None
        else:
            seed = SeedSequence(options.rng)

        log.info("training pipeline components")
        for node in self.nodes():
            match node:
                case ComponentInstanceNode(name, comp):
                    clog = log.bind(name=name, component=comp)
                    if isinstance(comp, Trainable):
                        # spawn new seed if needed
                        c_opts = options if seed is None else replace(options, rng=seed.spawn(1)[0])
                        clog.debug("training component")
                        comp.train(data, c_opts)
                    else:
                        clog.debug("training not required")
                case _:
                    pass

    @overload
    def run(self, /, **kwargs: object) -> object: ...
    @overload
    def run(self, node: str, /, **kwargs: object) -> object: ...
    @overload
    def run(self, nodes: tuple[str, ...], /, **kwargs: object) -> tuple[object, ...]: ...
    @overload
    def run(self, node: Node[T], /, **kwargs: object) -> T: ...
    @overload
    def run(self, nodes: tuple[Node[T1], Node[T2]], /, **kwargs: object) -> tuple[T1, T2]: ...
    @overload
    def run(
        self, nodes: tuple[Node[T1], Node[T2], Node[T3]], /, **kwargs: object
    ) -> tuple[T1, T2, T3]: ...
    @overload
    def run(
        self, nodes: tuple[Node[T1], Node[T2], Node[T3], Node[T4]], /, **kwargs: object
    ) -> tuple[T1, T2, T3, T4]: ...
    @overload
    def run(
        self,
        nodes: tuple[Node[T1], Node[T2], Node[T3], Node[T4], Node[T5]],
        /,
        **kwargs: object,
    ) -> tuple[T1, T2, T3, T4, T5]: ...
    def run(
        self,
        nodes: str | Node[Any] | tuple[str, ...] | tuple[Node[Any], ...] | None = None,
        /,
        **kwargs: object,
    ) -> object:
        """
        Run the pipeline and obtain the return value(s) of one or more of its
        components.  See :ref:`pipeline-execution` for details of the pipeline
        execution model.

        Args:
            nodes:
                The component(s) to run.
            kwargs:
                The pipeline's inputs, as defined with :meth:`create_input`.
                These are passed as-is to :meth:`run_all`, so they can also
                contain auxillary options like `_profile`.

        Returns:
            The pipeline result.  If no nodes are supplied, this is the result
            of the default node.  If a single node is supplied, it is the result
            of that node. If a tuple of nodes is supplied, it is a tuple of
            their results.

        Raises:
            PipelineError:
                when there is a pipeline configuration error (e.g. a cycle).
            ValueError:
                when one or more required inputs are missing.
            TypeError:
                when one or more required inputs has an incompatible type.
            other:
                exceptions thrown by components are passed through.
        """
        if nodes is None:
            if self._default:
                node_list = [self._default]
            else:
                raise RuntimeError("no node specified and pipeline has no default")
        elif isinstance(nodes, str) or isinstance(nodes, Node):
            node_list = [nodes]
        else:
            node_list = nodes

        state = self.run_all(*node_list, **kwargs)  # type: ignore
        results = [state[self.node(n).name] for n in node_list]

        if node_list is nodes:
            return tuple(results)
        else:
            return results[0]

    def run_all(
        self, *nodes: str | Node[Any], _profile: ProfileSink | None = None, **kwargs: object
    ) -> PipelineState:
        """
        Run all nodes in the pipeline, or all nodes required to fulfill the
        requested node, and return a mapping with the full pipeline state (the
        data attached to each node). This is useful in cases where client code
        needs to be able to inspect the data at arbitrary steps of the pipeline.
        It differs from :meth:`run` in two ways:

        1.  It returns the data from all nodes as a mapping (dictionary-like
            object), not just the specified nodes as a tuple.
        2.  If no nodes are specified, it runs *all* nodes.  This has the
            consequence of running nodes that are not required to fulfill the
            last node (such scenarios typically result from using
            :meth:`use_first_of`).

        Args:
            nodes:
                The nodes to run, as positional arguments (if no nodes are
                specified, this method runs all nodes).
            _profile:
                A profiler to profile this pipeline run.
            kwargs:
                The pipeline inputs.

        Returns:
            The full pipeline state, with :attr:`~PipelineState.default` set to
            the last node specified.
        """
        from .runner import PipelineRunner

        prof = _profile.run_profiler() if _profile is not None else None
        runner = PipelineRunner(self, kwargs, profiler=prof)
        node_list = [self.node(n) for n in nodes]
        _log.debug("running pipeline", name=self.name, nodes=[n.name for n in node_list])
        if not node_list:
            node_list = self.nodes()

        last = None
        for node in node_list:
            runner.run(node)
            last = node.name

        if prof is not None:
            prof.finish_run()

        return PipelineState(
            runner.state,
            {a: t.name for (a, t) in self._aliases.items()},
            default=last,
            meta=self.meta(),
        )

    def _check_member_node(self, node: Node[Any]) -> None:
        nw = self._nodes.get(node.name)
        if nw is not node:
            raise PipelineError(f"node {node} not in pipeline")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Literal, Mapping, TypeAlias, cast

import numpy as np
import structlog
from pydantic import AliasChoices, BaseModel, Field, PositiveFloat, PositiveInt
from scipy.sparse import coo_array
from typing_extensions import Generic, NamedTuple, TypeVar, override

from lenskit.config.common import EmbeddingSizeMixin
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.data.matrix import SparseRowArray
from lenskit.data.types import NPMatrix, NPVector, UIPair
from lenskit.logging import get_logger
from lenskit.parallel import ensure_parallel_init
from lenskit.pipeline import Component
from lenskit.training import ModelTrainer, TrainingOptions, UsesTrainer

_log = get_logger(__name__)

EntityClass: TypeAlias = Literal["user", "item"]

Scorer = TypeVar("Scorer", bound="ALSBase")
Config = TypeVar("Config", bound="ALSConfig")


class ALSConfig(EmbeddingSizeMixin, BaseModel):
    """
    Configuration for ALS scorers.
    """

    embedding_size: PositiveInt = Field(
        default=50, validation_alias=AliasChoices("embedding_size", "features")
    )
    """
    The dimension of user and item embeddings (number of latent features to
    learn).
    """
    epochs: PositiveInt = 10
    """
    The number of epochs to train.
    """
    regularization: PositiveFloat | UIPair[PositiveFloat] = 0.1
    """
    L2 regularization strength.
    """
    user_embeddings: bool | Literal["prefer"] = True
    """
    Whether to retain user embeddings after training.  If ``True``, they are
    retained, but are ignored if the query has historical items; if ``False``,
    they are not. If set to ``"prefer"``, then the user embeddings from training
    time are used even if the query has a user history.  This makes inference
    faster when histories only consist of the user's items from the training
    set.
    """

    @property
    def user_reg(self) -> float:
        if isinstance(self.regularization, UIPair):
            return self.regularization.user
        else:
            return self.regularization

    @property
    def item_reg(self) -> float:
        if isinstance(self.regularization, UIPair):
            return self.regularization.item
        else:
            return self.regularization


class TrainContext(NamedTuple):
    """
    Context object for one half of an ALS training operation.
    """

    label: str
    matrix: SparseRowArray
    left: NPMatrix
    right: NPMatrix
    reg: float
    nrows: int
    ncols: int
    embed_size: int
    regI: NPMatrix

    @classmethod
    def create(
        cls,
        label: str,
        matrix: SparseRowArray,
        left: NPMatrix,
        right: NPMatrix,
        reg: float,
    ) -> TrainContext:
        nrows, ncols = matrix.shape
        lnr, embed_size = left.shape
        assert lnr == nrows
        assert right.shape == (ncols, embed_size)
        regI = np.eye(embed_size, dtype=left.dtype) * reg
        return TrainContext(label, matrix, left, right, reg, nrows, ncols, embed_size, regI)


class ALSBase(UsesTrainer, Component[ItemList], ABC):
    """
    Base class for ALS models.

    Stability:
        Caller
    """

    config: ALSConfig

    users: Vocabulary | None
    items: Vocabulary
    user_embeddings: NPMatrix | None
    item_embeddings: NPMatrix

    @property
    def logger(self) -> structlog.stdlib.BoundLogger:
        return _log.bind(scorer=self.__class__.__name__, size=self.config.embedding_size)

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_id = query.user_id
        user_num = None
        if user_id is not None and self.users is not None:
            user_num = self.users.number(user_id, missing=None)

        log = self.logger.bind(user=user_id)

        u_offset = None
        u_feat = None
        if (
            query.user_items is not None
            and len(query.user_items) > 0
            and self.config.user_embeddings != "prefer"
        ):
            log.debug("training user embedding")
            u_feat, u_offset = self.new_user_embedding(user_num, query.user_items)

        if u_feat is None:
            if user_num is None or self.user_embeddings is None:
                log.debug("cannot find user embedding")
                return ItemList(items, scores=np.nan)
            u_feat = self.user_embeddings[user_num, :]

        item_nums = items.numbers(vocabulary=self.items, missing="negative")
        item_mask = item_nums >= 0

        scores = np.full((len(items),), np.nan, dtype=np.float32)
        if len(item_nums) <= self.item_embeddings.shape[0] * 0.8:
            # small set — subset inputs
            i_feats = self.item_embeddings[item_nums[item_mask], :]
            scores[item_mask] = i_feats @ u_feat
        else:
            # larger set — subset outputs
            allmult = self.item_embeddings @ u_feat
            scores[item_mask] = allmult[item_nums[item_mask]]

        log.debug("scored %d items", np.sum(item_mask))

        results = ItemList(items, scores=scores)
        return self.finalize_scores(user_num, results, u_offset)

    @abstractmethod
    def new_user_embedding(
        self, user_num: int | None, items: ItemList
    ) -> tuple[NPVector[np.float32], float | None]:  # pragma: no cover
        """
        Generate an embedding for a user given their current ratings.
        """
        ...

    def finalize_scores(
        self, user_num: int | None, items: ItemList, user_bias: float | None
    ) -> ItemList:
        """
        Perform any final transformation of scores prior to returning them.
        """
        return items


class ALSTrainerBase(ModelTrainer, Generic[Scorer, Config]):
    scorer: Scorer

    rng: np.random.Generator
    ui_rates: SparseRowArray
    "User-item rating matrix."
    u_ctx: TrainContext
    "User training context."
    iu_rates: SparseRowArray
    "Item-user rating matrix."
    i_ctx: TrainContext
    "Item training context."
    epochs_trained: int = 0

    def __init__(self, scorer: Scorer, data: Dataset, options: TrainingOptions):
        ensure_parallel_init()
        self.scorer = scorer
        self.scorer.users = data.users
        self.scorer.items = data.items

        self.rng = options.random_generator()

        ui_rates = self.prepare_matrix(data)
        self.ui_rates = SparseRowArray.from_scipy(ui_rates)
        self.iu_rates = SparseRowArray.from_scipy(ui_rates.T)

        self.initialize_params(data)
        self._init_contexts()

    def _init_contexts(self):
        assert self.scorer.user_embeddings is not None
        self.u_ctx = TrainContext.create(
            "user",
            self.ui_rates,
            self.scorer.user_embeddings,
            self.scorer.item_embeddings,
            self.config.user_reg,
        )
        self.i_ctx = TrainContext.create(
            "item",
            self.iu_rates,
            self.scorer.item_embeddings,
            self.scorer.user_embeddings,
            self.config.item_reg,
        )

    def train_epoch(self):
        epoch = self.epochs_trained + 1
        log = self.logger.bind(epoch=epoch)

        assert self.scorer.user_embeddings is not None
        assert self.scorer.item_embeddings is not None

        du = self.als_half_epoch(epoch, self.u_ctx)
        log.debug("finished user epoch")

        di = self.als_half_epoch(epoch, self.i_ctx)
        log.debug("finished item epoch")

        log.debug("finished epoch (|ΔP|=%.3f, |ΔQ|=%.3f)", du, di)
        self.epochs_trained += 1
        return {"deltaP": du, "deltaQ": di}

    @property
    def config(self) -> Config:
        return cast(Config, self.scorer.config)

    @property
    def logger(self) -> structlog.stdlib.BoundLogger:
        return self.scorer.logger

    @property
    def n_users(self):
        return self.ui_rates.shape[0]

    @property
    def n_items(self):
        return self.iu_rates.shape[1]

    @abstractmethod
    def prepare_matrix(self, data: Dataset) -> coo_array:  # pragma: no cover
        """
        Prepare data for training this model.  This takes in the ratings, and is
        supposed to do two things:

        -   Normalize or transform the rating/interaction data, as needed, for
            training.
        -   Store any parameters learned from the normalization (e.g. means) in
            the appropriate member variables.
        -   Return the ratings matrix for training.
        """

    def initialize_params(self, data: Dataset):
        """
        Initialize the model parameters at the beginning of training.
        """
        self.logger.debug("initializing item matrix")
        self.scorer.item_embeddings = self.initial_params(
            data.item_count, self.config.embedding_size
        )
        self.logger.debug("|Q|: %f", np.linalg.norm(self.scorer.item_embeddings, "fro"))

        self.logger.debug("initializing user matrix")
        self.scorer.user_embeddings = self.initial_params(
            data.user_count, self.config.embedding_size
        )
        self.logger.debug("|P|: %f", np.linalg.norm(self.scorer.user_embeddings, "fro"))

    @abstractmethod
    def initial_params(self, nrows: int, ncols: int) -> NPMatrix:  # pragma: no cover
        """
        Compute initial parameter values of the specified shape.
        """
        ...

    @abstractmethod
    def als_half_epoch(self, epoch: int, context: TrainContext) -> float:  # pragma: no cover
        """
        Run one half of an ALS training epoch.
        """
        ...

    @override
    def finalize(self):
        """
        Finalize training. Base classes must call superclass.
        """
        self.logger.debug("finalizing model training")
        if not self.config.user_embeddings:
            self.scorer.user_embeddings = None
            self.scorer.users = None

    def get_parameters(self) -> Mapping[str, object]:
        """
        Get the component's parameters.

        Returns:
            The model's parameters, as a dictionary from names to parameter data
            (usually arrays, tensors, etc.).
        """
        return {
            "user_embeddings": self.scorer.user_embeddings,
            "item_embeddings": self.scorer.item_embeddings,
        }

    def load_parameters(self, state: Mapping[str, object]) -> None:
        """
        Reload model state from parameters saved via :meth:`get_parameters`.

        Args:
            params:
                The model parameters, as a dictionary from names to parameter
                data (arrays, tensors, etc.), as returned from
                :meth:`get_parameters`.
        """
        u_emb = state["user_embeddings"]
        assert isinstance(u_emb, np.ndarray)
        i_emb = state["item_embeddings"]
        assert isinstance(i_emb, np.ndarray)
        self.scorer.user_embeddings = u_emb
        self.scorer.item_embeddings = i_emb
        self._init_contexts()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Bridges to recommendation models from :mod:`implicit`.
"""

import logging

import numpy as np
from implicit.als import AlternatingLeastSquares
from implicit.bpr import BayesianPersonalizedRanking
from implicit.recommender_base import RecommenderBase
from numpy.typing import NDArray
from pydantic import BaseModel, JsonValue
from scipy.sparse import csr_matrix
from typing_extensions import override

from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.parallel.config import ensure_parallel_init, get_parallel_config
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_logger = logging.getLogger(__name__)

__all__ = [
    "BaseRec",
    "ALS",
    "BPR",
]


class ImplicitConfig(BaseModel, extra="allow"):
    __pydantic_extra__: dict[str, JsonValue]


class ImplicitALSConfig(ImplicitConfig, extra="allow"):
    weight: float = 40.0


class BaseRec(Component[ItemList], Trainable):
    """
    Base class for Implicit-backed recommenders.

    Stability:
        Caller
    """

    config: ImplicitConfig
    weight: float = 1.0

    matrix: csr_matrix
    """
    The user-item rating matrix from training.
    """
    users: Vocabulary
    """
    The user ID mapping from training.
    """
    items: Vocabulary
    """
    The item ID mapping from training.
    """
    user_embeddings: NDArray[np.float32]
    item_embeddings: NDArray[np.float32]

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_embeddings") and not options.retrain:
            return

        matrix = data.interaction_matrix(format="scipy", layout="csr", legacy=True)
        uir = matrix * self.weight

        delegate = self._construct()
        _logger.info("training %s on %s matrix (%d nnz)", delegate, uir.shape, uir.nnz)

        delegate.fit(uir)

        self.matrix = matrix
        self.users = data.users
        self.items = data.items

        if hasattr(delegate.item_factors, "to_numpy"):
            self.user_embeddings = delegate.user_factors.to_numpy()
            self.item_embeddings = delegate.item_factors.to_numpy()
        else:
            self.user_embeddings = delegate.user_factors
            self.item_embeddings = delegate.item_factors

        return self

    def _construct(self) -> RecommenderBase:
        raise NotImplementedError("implicit constructor not implemented")

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_num = None
        if query.user_id is not None:
            user_num = self.users.number(query.user_id, missing=None)

        if user_num is None or not len(items):
            return ItemList(items, scores=np.nan)

        inos = items.numbers(vocabulary=self.items, missing="negative")
        mask = inos >= 0
        good_inos = inos[mask]
        mult_first = len(good_inos) > len(self.items) // 2

        if mult_first:
            ifs = self.item_embeddings
        else:
            ifs = self.item_embeddings[good_inos]
        uf = self.user_embeddings[user_num]

        prod = np.dot(ifs, uf.T).reshape(-1)
        if mult_first:
            prod = prod[good_inos]

        scores = np.full(len(items), np.nan)
        scores[mask] = prod

        return ItemList(items, scores=scores)


class ALS(BaseRec):
    """
    LensKit interface to :py:mod:`implicit.cpu.als` (or GPU version).

    Stability:
        Caller
    """

    config: ImplicitALSConfig

    @property
    def weight(self):
        return self.config.weight

    def _construct(self):
        ensure_parallel_init()
        pcfg = get_parallel_config()
        return AlternatingLeastSquares(
            num_threads=pcfg.threads,
            **self.config.__pydantic_extra__,  # type: ignore
        )


class BPR(BaseRec):
    """
    LensKit interface to :py:mod:`implicit.cpu.bpr` (or GPU version).

    Stability:
        Caller
    """

    def _construct(self):
        ensure_parallel_init()
        pcfg = get_parallel_config()
        return BayesianPersonalizedRanking(
            num_threads=pcfg.threads,
            **self.config.__pydantic_extra__,  # type: ignore
        )



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Callable, TypeVar

import numpy as np
import pandas as pd
import pyarrow as pa

from lenskit.data import ItemList, ItemListCollection

from ._base import DecomposedMetric, GlobalMetric, ListMetric, Metric, MetricFunction

_log = logging.getLogger(__name__)
K1 = TypeVar("K1", bound=tuple)
K2 = TypeVar("K2", bound=tuple)


@dataclass(frozen=True)
class MetricWrapper:
    """
    Internal class for storing metrics with their configuration.

    Stability:
        Internal
    """

    metric: Metric | MetricFunction
    label: str
    default: float | None = None

    @property
    def is_listwise(self) -> bool:
        "Check if this metric is listwise."
        return isinstance(self.metric, (ListMetric, Callable))

    @property
    def is_global(self) -> bool:
        "Check if this metric is global."
        return isinstance(self.metric, GlobalMetric)

    @property
    def is_decomposed(self) -> bool:
        "Check if this metric is decomposed."
        return isinstance(self.metric, DecomposedMetric)

    def measure_list(self, list: ItemList, test: ItemList) -> Any:
        """Get intermediate measurement data from the metric."""
        if isinstance(self.metric, Callable):
            return self.metric(list, test)
        return self.metric.measure_list(list, test)

    def extract_list_metrics(self, intermediate_data: Any) -> float | dict[str, float] | None:
        """Extract per-list metrics from intermediate data."""
        return self.metric.extract_list_metrics(intermediate_data)

    def summarize(
        self, values: list[Any] | pa.Array | pa.ChunkedArray
    ) -> dict[str, float | None] | float | None:
        """Aggregate intermediate values into summary statistics."""
        return self.metric.summarize(values)

    def measure_run(self, run: ItemListCollection, test: ItemListCollection) -> float:
        """Only global metrics support run-level measurement."""
        if isinstance(self.metric, GlobalMetric):
            return self.metric.measure_run(run, test)
        else:
            raise TypeError(f"metric {self.metric} does not support global measurement")


class MetricAccumulator:
    """
    Accumulates metric measurements over multiple recommendation lists.

    This class separates metric accumulation from the evaluation loop.

    Stability:
        Caller
    """

    def __init__(self):
        self.metrics: list[MetricWrapper] = []
        self._records: list[dict[str, Any]] = []
        self._key_fields: list[str] = []

    def add_metric(
        self,
        metric: Metric | MetricFunction | type[Metric],
        label: str | None = None,
        default: float | None = None,
    ):
        """
        Add a metric to this accumulator.

        Args:
            metric:
                The metric to add.
            label:
                The label to use for the metric's results. If unset, obtains
                from the metric.
            default:
                The default value to use when a user does not have
                recommendations. If unset, obtains from the metric's ``default``
                attribute (if specified).
        """
        wrapper = self._wrap_metric(metric, label, default)
        self.metrics.append(wrapper)

    def measure_list(self, output: ItemList, test: ItemList, **keys: Any):
        """
        Measure a single list and accumulate the intermediate results.

        Args:
            output:
                The recommendation list to measure.
            test:
                The ground truth test data.
            **keys:
                Identifying keys for this list (e.g., user_id).
        """
        if not self._key_fields:
            self._key_fields = list(keys.keys())

        rec = dict(keys)
        for wrapper in self.metrics:
            if wrapper.is_global:
                continue

            intermediate_data = wrapper.measure_list(output, test)

            rec[f"_intermediate_{wrapper.label}"] = intermediate_data

        self._records.append(rec)

    def list_metrics(self, fill_missing: bool = True) -> pd.DataFrame:
        """
        Get the per-list metric results as a DataFrame.

        Args:
            fill_missing: If True, fill missing values with metric defaults.

        Returns:
            DataFrame with one row per list and one column per metric.
        """
        if not self._records:
            return pd.DataFrame()

        extracted_records = []

        for record in self._records:
            new_record = {k: v for k, v in record.items() if not k.startswith("_intermediate_")}

            for wrapper in self.metrics:
                if wrapper.is_global:
                    continue

                intermediate_key = f"_intermediate_{wrapper.label}"
                if intermediate_key in record:
                    intermediate_data = record[intermediate_key]

                    extracted_val = wrapper.extract_list_metrics(intermediate_data)

                    if extracted_val is None and wrapper.default is not None:
                        extracted_val = wrapper.default

                    if isinstance(extracted_val, dict):
                        for k, v in extracted_val.items():
                            new_record[f"{wrapper.label}.{k}"] = v
                    else:
                        new_record[wrapper.label] = extracted_val

            extracted_records.append(new_record)

        df = pd.DataFrame(extracted_records)

        if self._key_fields:
            df.set_index(self._key_fields, inplace=True, drop=True)

        if fill_missing:
            for wrapper in self.metrics:
                if wrapper.default is not None and not wrapper.is_global:
                    label = wrapper.label
                    nested_cols = [c for c in df.columns if c.startswith(f"{label}.")]
                    if nested_cols:
                        for col in nested_cols:
                            df[col] = df[col].fillna(wrapper.default)
                    elif label in df.columns:
                        df[label] = df[label].fillna(wrapper.default)

        return df

    def summary_metrics(self) -> dict[str, float]:
        """
        Compute summary statistics by calling each metric's summarize() method.

        Returns:
            A dictionary with flattened metric results.
        """
        results = {}

        for wrapper in self.metrics:
            if wrapper.is_global:
                continue

            label = wrapper.label
            intermediate_key = f"_intermediate_{label}"

            intermediate_values = []
            for record in self._records:
                if intermediate_key in record and record[intermediate_key] is not None:
                    intermediate_values.append(record[intermediate_key])

            if intermediate_values:
                summary = wrapper.summarize(intermediate_values)
                if summary is not None:
                    if isinstance(summary, dict):
                        for key, value in summary.items():
                            results[f"{label}.{key}"] = value
                    else:
                        results[label] = summary

        return results

    def _wrap_metric(
        self,
        m: Metric | MetricFunction | type[Metric],
        label: str | None = None,
        default: float | None = None,
    ) -> MetricWrapper:
        """Wrap a metric with its configuration."""
        if isinstance(m, type):
            m = m()

        if label is None:
            if isinstance(m, Metric):
                wl = m.label
            elif isinstance(m, type):
                wl = m.__name__  # type: ignore
            else:
                wl = type(m).__name__
        else:
            wl = label

        if default is None:
            if isinstance(m, ListMetric):
                default = m.default

            if default is not None and not isinstance(
                default, (float, int, np.floating, np.integer)
            ):
                raise TypeError(f"metric {m} has unsupported default {default}")

        return MetricWrapper(m, wl, default)  # type: ignore

    def _validate_setup(self):
        """Validate that the accumulator is properly configured."""
        seen = set()
        for m in self.metrics:
            lbl = m.label
            if lbl in seen:
                raise RuntimeError(f"duplicate metric: {lbl}")
            seen.add(lbl)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Any, Literal, Mapping, TypeAlias

import pandas as pd

from lenskit.data import ID, GenericKey, ItemList, ItemListCollection, UserIDKey
from lenskit.logging import Stopwatch, get_logger, item_progress
from lenskit.parallel import invoker
from lenskit.pipeline import Pipeline, PipelineProfiler
from lenskit.pipeline._profiling import ProfileSink

from ._results import BatchResults

_log = get_logger(__name__)

ItemSource: TypeAlias = None | Literal["test-items"]
"""
Types of items that can be returned.
"""


@dataclass
class InvocationSpec:
    """
    Specification for a single pipeline invocation, to record one or more
    pipeline component outputs for a test user.
    """

    name: str
    "A name for this invocation."
    components: dict[str, str]
    "The names of pipeline components to measure and return, mapped to their output names."
    items: ItemSource = None
    "The target or candidate items (if any) to provide to the recommender."
    extra_inputs: dict[str, Any] = field(default_factory=dict)
    "Additional inputs to pass to the pipeline."


class BatchPipelineRunner:
    """
    Apply a pipeline to a collection of test users.

    Stability:
        Caller

    Argss:
        pipeline:
            The pipeline to evaluate.
        n_jobs:
            The number of parallel processes to use, or ``None`` for the
            default (defined by :func:`lenskit.parallel.config.initialize`).
    """

    n_jobs: int | Literal["ray"] | None
    profiler: PipelineProfiler | None
    invocations: list[InvocationSpec]

    def __init__(
        self,
        *,
        n_jobs: int | Literal["ray"] | None = None,
        profiler: PipelineProfiler | None = None,
    ):
        self.n_jobs = n_jobs
        self.profiler = profiler
        self.invocations = []

    def add_invocation(self, inv: InvocationSpec):
        self.invocations.append(inv)

    def score(self, component: str = "scorer", *, output: str = "scores"):
        """
        Request the batch run to generate test item scores.

        Args:
            component:
                The name of the rating predictor component to run.
            output:
                The name of the results in the output dictionary.
        """
        self.add_invocation(InvocationSpec("score", {component: output}, "test-items"))

    def predict(self, component: str = "rating-predictor", *, output: str = "predictions"):
        """
        Request the batch run to generate test item rating predictions.  It is identical
        to :meth:`score` but with different defaults.

        Args:
            component:
                The name of the rating predictor component to run.
            output:
                The name of the results in the output dictionary.
        """
        return self.score(component, output=output)

    def recommend(
        self, component: str = "recommender", *, output: str = "recommendations", **extra: Any
    ):
        """
        Request the batch run to generate recomendations.

        Args:
            component:
                The name of the recommender component to run.
            output:
                The name of the results in the output dictionary.
            extra:
                Extra inputs to the recommender. A common option is ``n``, the
                number of recommendations to return (a default may be baked into
                the pipeline).
        """
        self.add_invocation(InvocationSpec("recommend", {component: output}, extra_inputs=extra))

    def run(
        self,
        pipeline: Pipeline,
        test_data: ItemListCollection[GenericKey]
        | Mapping[ID, ItemList]
        | Iterable[ID | GenericKey]
        | pd.DataFrame,
    ) -> BatchResults:
        """
        Run the pipeline and return its results.

        Args:
            test_data:
                The collection of test data, as an ItemListCollection, a mapping
                of user IDs to test data, or as a sequence of item IDs for
                recommendation.

        Returns:
            The results, as a nested dictionary.  The outer dictionary maps
            component output names to inner dictionaries of result data.  These
            inner dictionaries map user IDs to
        """
        if isinstance(test_data, pd.DataFrame):
            test_data = ItemListCollection.from_df(test_data)

        if isinstance(test_data, ItemListCollection):
            test_iter = test_data.items()
            key_type = test_data.key_type
            n_users = len(test_data)
        elif isinstance(test_data, Mapping):
            key_type = UserIDKey
            test_iter = ((UserIDKey(k), v) for (k, v) in test_data.items())  # type: ignore
            n_users = len(test_data)
        else:
            key_type = UserIDKey
            test_data = list(test_data)
            test_iter = ((_ensure_key(k), None) for k in test_data)
            n_users = len(test_data)

        prof = self.profiler
        if prof is not None:
            prof = prof.multiprocess()

        log = _log.bind(name=pipeline.name, n_queries=n_users, n_jobs=self.n_jobs)
        log.info("beginning batch run")

        with (
            invoker(
                (pipeline, self.invocations, prof), _run_pipeline, n_jobs=self.n_jobs
            ) as worker,
            item_progress("Recommending", n_users) as progress,
        ):
            # release our reference, will sometimes free the pipeline memory in this process
            del pipeline
            results = BatchResults(key_type)
            timer = Stopwatch()
            for key, outs in worker.map(test_iter):
                for cn, cr in outs.items():
                    results.add_result(cn, key, cr)
                progress.update()
            timer.stop()

            rate_ms = timer.elapsed() / n_users * 1000
            log.info("finished running in %s", timer, time_per_query="{:.1f}ms".format(rate_ms))

        return results


def _ensure_key(key: ID | tuple[ID, ...]):
    if isinstance(key, tuple):
        return key
    else:
        return UserIDKey(key)


def _run_pipeline(
    ctx: tuple[Pipeline, list[InvocationSpec], ProfileSink | None],
    req: tuple[GenericKey, ItemList],
) -> tuple[GenericKey, dict[str, object]]:
    pipeline, invocations, profiler = ctx
    key, test_items = req

    result = {}

    _log.debug("running pipeline", name=pipeline.name, key=key)
    for inv in invocations:
        inputs: dict[str, Any] = {}
        if hasattr(key, "user_id"):
            inputs["query"] = key.user_id  # type: ignore
        match inv.items:
            case "test-items":
                inputs["items"] = test_items

        inputs.update(inv.extra_inputs)

        nodes = inv.components.keys()
        outs = pipeline.run_all(*nodes, _profile=profiler, **inputs)
        for cname, oname in inv.components.items():
            result[oname] = outs[cname]

    return key, result  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np

from lenskit.data import ItemList
from lenskit.metrics.reranking import least_item_promoted


def test_lip_empty_reference():
    ref = ItemList()
    rerank = ItemList([1, 2, 3], ordered=True)
    assert np.isnan(least_item_promoted(ref, rerank))


def test_lip_no_overlap():
    ref = ItemList([10, 11, 12], ordered=True)
    rerank = ItemList([1, 2, 3], ordered=True)
    # none of the reranked items exist in reference
    # lip_rank stays at k, so result = 0
    assert least_item_promoted(ref, rerank, k=3) == 0


def test_lip_perfect_alignment():
    ref = ItemList([1, 2, 3, 4, 5], ordered=True)
    rerank = ItemList([1, 2, 3, 4, 5], ordered=True)
    # top-k reranked items are already top-k in reference
    assert least_item_promoted(ref, rerank, k=3) == 0


def test_lip_promoted_from_below_k():
    ref = ItemList([1, 2, 3, 4, 5, 6], ordered=True)
    rerank = ItemList([6, 2, 3, 1], ordered=True)
    # item 6 is at position 5 in reference but appears in reranked top-3
    # lip_rank = 5, so return = 5 - 3 = 2
    assert least_item_promoted(ref, rerank, k=3) == 2


def test_lip_multiple_promotions():
    ref = ItemList([1, 2, 3, 4, 5, 6, 7], ordered=True)
    rerank = ItemList([7, 6, 2, 3], ordered=True)
    # item 7 is at rank 6, promoted into top-3
    # lip_rank = 6, so result = 6 - 3 = 3
    assert least_item_promoted(ref, rerank, k=3) == 3



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from itertools import product

from pytest import mark

from lenskit.flexmf import FlexMFImplicitConfig, FlexMFImplicitScorer
from lenskit.testing import BasicComponentTests, ScorerTests


class TestFlexMFImplicit(BasicComponentTests, ScorerTests):
    expected_ndcg = (0.01, 0.25)
    component = FlexMFImplicitScorer
    config = FlexMFImplicitConfig()


class TestFlexMFWARP(BasicComponentTests, ScorerTests):
    component = FlexMFImplicitScorer
    config = FlexMFImplicitConfig(loss="warp")


def test_config_defaults():
    cfg = FlexMFImplicitConfig()
    assert cfg.embedding_size == 64


def test_config_exp_ctor():
    cfg = FlexMFImplicitConfig(embedding_size_exp=5)  # type: ignore
    assert cfg.embedding_size == 32


def test_config_exp_dict():
    cfg = FlexMFImplicitConfig.model_validate({"embedding_size_exp": 10})
    assert cfg.embedding_size == 1024


def test_config_exp_json():
    cfg = FlexMFImplicitConfig.model_validate_json('{"embedding_size_exp": 2}')
    assert cfg.embedding_size == 4


@mark.slow
@mark.parametrize(["loss", "reg"], product(["logistic", "pairwise"], ["L2", "AdamW"]))
def test_flexmf_train_config(ml_ds, loss, reg):
    config = FlexMFImplicitConfig(loss=loss, reg_method=reg)
    model = FlexMFImplicitScorer(config)
    print("training", model)
    model.train(ml_ds)

    assert model.model is not None



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import numpy as np
from scipy.sparse import coo_array
from typing_extensions import override

from lenskit._accel import als
from lenskit.data import Dataset, ItemList
from lenskit.data.types import NPMatrix, NPVector
from lenskit.logging import item_progress
from lenskit.math.solve import solve_cholesky

from ._common import ALSBase, ALSConfig, ALSTrainerBase, TrainContext


class ImplicitMFConfig(ALSConfig):
    weight: float = 40
    """
    The confidence weight for positive examples.
    """
    use_ratings: bool = False
    """
    If ``True``, use rating values instead of just presence or absence.
    """


class ImplicitMFScorer(ALSBase):
    """
    Implicit matrix factorization trained with alternating least squares
    :cite:p:`hu:implicit-mf`.  This algorithm outputs 'predictions', but they
    are not on a meaningful scale.  If its input data contains ``rating``
    values, these will be used as the 'confidence' values; otherwise, confidence
    will be 1 for every rated item.

    With weight :math:`w`, this function decomposes the matrix
    :math:`\\mathbb{1}^* + Rw`, where :math:`\\mathbb{1}^*` is an :math:`m
    \\times n` matrix of all 1s.

    See the base class :class:`ALSBase` for documentation on the estimated
    parameters you can extract from a trained model. See
    :class:`ImplicitMFConfig` and :class:`ALSConfig` for the configuration
    options for this component.

    .. versionchanged:: 2025.1
        ``ImplicitMFScorer`` no longer supports multiple training methods. It always uses
        Cholesky decomposition now.

    .. versionchanged:: 0.14
        By default, ``ImplicitMF`` ignores a ``rating`` column if one is present in the training
        data.  This can be changed through the ``use_ratings`` option.

    .. versionchanged:: 0.13
        In versions prior to 0.13, ``ImplicitMF`` used the rating column if it was present.
        In 0.13, we added an option to control whether or not the rating column is used; it
        initially defaulted to ``True``, but with a warning.  In 0.14 it defaults to ``False``.

    Stability:
        Caller
    """

    config: ImplicitMFConfig

    _OtOr: NPMatrix

    def create_trainer(self, data, options):
        return ImplicitMFTrainer(self, data, options)

    @override
    def new_user_embedding(
        self, user_num: int | None, user_items: ItemList
    ) -> tuple[NPVector, None]:
        ri_idxes = user_items.numbers(vocabulary=self.items)

        ri_good = ri_idxes >= 0
        ri_it = ri_idxes[ri_good]
        if self.config.use_ratings:
            ratings = user_items.field("rating")
            if ratings is None:
                raise ValueError("no ratings in user items")
            ri_val = ratings[ri_good] * self.config.weight
        else:
            ri_val = np.full((len(ri_good),), self.config.weight)

        ri_val = ri_val.astype(self.item_embeddings.dtype)

        u_feat = self._train_new_row(ri_it, ri_val, self.item_embeddings, self._OtOr)
        return u_feat, None

    def _train_new_row(
        self,
        items: NPVector[np.int32],
        ratings: NPVector,
        i_embeds: NPMatrix,
        OtOr: NPMatrix,
    ) -> NPVector:
        """
        Train a single user row with new rating data.

        Args:
            items: the item IDs the user has rated
            ratings: the user's ratings for those items (when rating values are used)
            other: the item-feature matrix
            OtOr: the pre-computed regularization and background matrix.

        Returns:
            The user-feature vector.
        """
        self.logger.debug("learning new user row", n_items=len(items))

        # we can optimize by only considering the nonzero entries of Cu-I
        # this means we only need the corresponding matrix columns
        M = i_embeds[items, :]
        # Compute M^T (C_u-I) M, restricted to these nonzero entries
        MMT = (M.T * ratings) @ M
        # Build the matrix for solving
        A = OtOr + MMT
        # Compute RHS - only used columns (p_ui != 0) values needed
        y = i_embeds.T[:, items] @ (ratings + 1.0)
        # and solve
        x = solve_cholesky(A, y)

        return x


class ImplicitMFTrainer(ALSTrainerBase[ImplicitMFScorer, ImplicitMFConfig]):
    @override
    def train_epoch(self):
        res = super().train_epoch()
        self._save_user_otor()
        return res

    @override
    def prepare_matrix(self, data: Dataset) -> coo_array:
        ints = data.interactions().matrix()
        if self.config.use_ratings:
            rmat = ints.scipy(attribute="rating", layout="coo")
        else:
            rmat = ints.scipy(layout="coo")

        vals = np.require(rmat.data, dtype=np.float32) * self.config.weight
        return coo_array((vals, (rmat.row, rmat.col)), shape=rmat.shape)

    @override
    def initial_params(self, nrows: int, ncols: int) -> NPMatrix:
        mat = self.rng.standard_normal((nrows, ncols), dtype=np.float32) * 0.01
        mat *= mat
        return mat

    @override
    def als_half_epoch(self, epoch: int, context: TrainContext) -> float:
        OtOr = _implicit_otor(context.right, context.reg)
        with item_progress(f"Epoch {epoch} {context.label}", context.nrows) as pb:
            return als.train_implicit_matrix(context.matrix, context.left, context.right, OtOr, pb)

    @override
    def finalize(self):
        self._save_user_otor()
        super().finalize()

    def _save_user_otor(self):
        # compute OtOr and save it on the model
        reg = self.config.user_reg
        self.scorer._OtOr = _implicit_otor(self.scorer.item_embeddings, reg)


def _implicit_otor(other: NPMatrix, reg: float) -> NPMatrix:
    nf = other.shape[1]
    regmat = np.eye(nf, dtype=other.dtype)
    regmat *= reg
    Ot = other.T
    OtO = Ot @ other
    OtO += regmat
    return OtO



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import os
from contextlib import contextmanager
from pathlib import Path
from shutil import copyfileobj
from urllib.parse import urlparse
from urllib.request import urlopen

import click
from humanize import naturalsize

from lenskit.logging import get_logger

_log = get_logger(__name__)
ML_LOC = "http://files.grouplens.org/datasets/movielens/"
AZ_LOC = {
    "2023": "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/benchmark/0core/rating_only/{}.csv.gz",
    "2023-5core": "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_2023/benchmark/5core/rating_only/{}.csv.gz",
    "2018": "https://mcauleylab.ucsd.edu/public_datasets/data/amazon_v2/categoryFilesSmall/{}.csv",
    "2014": "https://snap.stanford.edu/data/amazon/productGraph/categoryFiles/ratings_{}.csv",
}


@click.command("fetch")
@click.option("--movielens", "source", flag_value="movielens", help="Fetch MovieLens data.")
@click.option("--amazon", "source", flag_value="amazon", help="Fetch Amazon ratings data.")
@click.option("--edition", default="2023", help="Amazon ratings edition.")
@click.option("--core", is_flag=True, help="Fetch core instead of full data.")
@click.option("--force", is_flag=True, help="overwrite existing file")
@click.option("-D", "--data-dir", "dest", type=Path, help="directory for downloaded data")
@click.argument("name", nargs=-1)
def fetch(
    source: str | None, dest: Path | None, name: list[str], force: bool, edition: str, core: bool
):
    """
    Convert data into the LensKit native format.
    """

    if dest is None:
        dest = Path()

    dest.mkdir(exist_ok=True, parents=True)

    match source:
        case None:
            _log.error("no data source specified")
            raise click.UsageError("no data source specified")
        case "movielens":
            for n in name:
                fetch_movielens(n, dest, force)
        case "amazon":
            for n in name:
                fetch_amazon_ratings(edition, core, n, dest, force)
        case _:
            raise ValueError(f"unknown data format {source}")


def fetch_movielens(name: str, base_dir: Path, force: bool):
    """
    Fetch a MovieLens dataset.  The followings names are recognized:

    . ml-100k
    . ml-1m
    . ml-10m
    . ml-20m
    . ml-25m
    . ml-latest
    . ml-latest-small

    Args:
        name:
            The name of the dataset.
        base_dir:
            The base directory into which data should be downloaded.
    """
    zipname = f"{name}.zip"
    zipfile = base_dir / zipname
    zipurl = ML_LOC + zipname

    log = _log.bind(source="movielens", name=name, dest=str(zipfile))

    if zipfile.exists():
        if force:
            log.warning("output file already exists, ovewriting")
        else:
            log.info("output file already exists")
            return

    log.debug("ensuring parent directory exists")

    log.info("downloading MovieLens data set")
    with output_file(zipfile) as zf:
        res = urlopen(zipurl)
        copyfileobj(res, zf, 8 * 1024 * 1024)

    log.info("downloaded %s", naturalsize(zipfile.stat().st_size, binary=True))


def fetch_amazon_ratings(edition: str, core: bool, name: str, base_dir: Path, force: bool):
    key = edition + "-5core" if core else edition
    log = _log.bind(source="amazon", name=name, edition=edition, core=core)
    pat = AZ_LOC[key]

    url = pat.format(name)
    fn = os.path.basename(urlparse(url).path)

    out_path = base_dir / fn
    if out_path.exists() and not force:
        log.info("output file already exists")
        return

    log.info("downloading %s", fn)
    with output_file(out_path) as outf:
        res = urlopen(url)
        copyfileobj(res, outf, 8 * 1024 * 1024)

    log.info("downloaded %s", naturalsize(out_path.stat().st_size, binary=True))


@contextmanager
def output_file(path: Path):
    """
    Open an output file and write.
    """
    pid = os.getpid()
    tmp = path.with_name(f".{path.name}.{pid}.tmp")
    _log.debug("opening temp file", name=tmp.name)
    try:
        with open(tmp, "wb") as file:
            yield file
        _log.debug("replacing real file", name=path.name)
        os.replace(tmp, path)
    finally:
        if tmp.exists():
            _log.debug("write failed, deleting temp file", name=tmp.name)
            tmp.unlink()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Vocabularies of IDs, tags, etc.
"""

# pyright: basic
from __future__ import annotations

from typing import Hashable, Iterable, Iterator, Literal, Sequence, overload

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from numpy.typing import ArrayLike, NDArray
from structlog.stdlib import BoundLogger

from lenskit import _accel
from lenskit.logging import get_logger, trace

from .types import ID, IDArray, IDSequence

_log = get_logger(__name__)


class Vocabulary:
    """
    Vocabularies of entity identifiers for the LensKit data model.

    This class supports bidirectional mappings between key-like data and
    congiguous nonnegative integer indices.  Its key use is to facilitate the
    entity ID vocabularies in :class:`~lenskit.data.Dataset`, but it can also be
    used for things like item tags.

    IDs in a vocabulary must be unique.  Constructing a vocabulary with
    ``reorder=True`` ensures uniqueness (and sorts the IDs), but does not
    preserve the order of IDs in the original input.

    It is currently a wrapper around :class:`pandas.Index`, but this fact is not
    part of the stable public API.

    Args:
        keys:
            The IDs to put in the vocabulary.
        name:
            The vocabulary name (i.e. the entity class it stores IDs for).
        reorder:
            If ``True``, sort and deduplicate the IDs.  If ``False`` (the
            default), use the IDs as-is (assigning each to their position in the
            input sequence).

    Stability:
        Caller
    """

    name: str | None
    "The name of the vocabulary (e.g. “user”, “item”)."
    _pd_index: pd.Index | None = None
    "The Pandas index implementing the vocabulary."
    _array: pa.Array
    "PyArrow array for the vocabulary."
    _index: _accel.data.IDIndex
    "Internal index."
    _hash: str | None
    "Checksum of index data for fast equivalence testing."
    _log: BoundLogger

    def __init__(
        self,
        keys: IDSequence | pd.Index | Iterable[ID] | None = None,
        name: str | None = None,
        *,
        reorder: bool = True,
    ):
        self.name = name
        self._log = _log.bind(entity=name)
        key_arr: pa.Array
        if keys is None:
            key_arr = pa.array([], type=pa.int32())
        elif isinstance(keys, pd.Index):
            self._pd_index = keys
            key_arr = pa.array(keys.values)
        elif isinstance(keys, np.ndarray) or isinstance(keys, list) or isinstance(keys, pd.Series):
            key_arr = pa.array(keys)  # type: ignore
        elif isinstance(keys, pa.ChunkedArray):
            key_arr = keys.drop_null().combine_chunks()
        elif isinstance(keys, pa.Array):
            key_arr = keys
        else:
            key_arr = pa.array(sorted(set(keys)))  # type: ignore

        if reorder:
            self._pd_index = None
            key_arr = key_arr.unique().sort()

        assert key_arr.null_count == 0
        self._array = key_arr
        self._index = _accel.data.IDIndex(key_arr)
        self._ensure_hash()

    def _ensure_hash(self):
        if not hasattr(self, "_hash"):
            # since we just made this array, we can assume the buffer is fully used
            self._hash = _accel.data.hash_array(self._array)

    @property
    def index(self) -> pd.Index:
        """
        The vocabulary as a Pandas index.

        Stability:
            Internal
        """
        if self._pd_index is None:
            self._pd_index = pd.Index(self._array.to_numpy(zero_copy_only=False))
            self._pd_index.name = f"{self.name}_id"
        return self._pd_index

    @property
    def size(self) -> int:
        "Current vocabulary size."
        return len(self._array)

    @overload
    def number(self, term: object, missing: Literal["error"] = "error") -> int: ...
    @overload
    def number(self, term: object, missing: Literal["none"] | None) -> int | None: ...
    def number(
        self, term: object, missing: Literal["error", "none"] | None = "error"
    ) -> int | None:
        "Look up the number for a vocabulary ID."
        idx = self._index.get_index(term)  # type: ignore
        if idx is None and missing == "error":
            raise KeyError(f"{self.name} ID {term}")

        return idx

    @overload
    def numbers(
        self,
        terms: Sequence[Hashable] | ArrayLike,
        missing: Literal["error", "negative"] = "error",
        *,
        format: Literal["numpy"] = "numpy",
    ) -> NDArray[np.int32]: ...
    @overload
    def numbers(
        self,
        terms: Sequence[Hashable] | ArrayLike,
        missing: Literal["error", "negative", "null"] = "error",
        *,
        format: Literal["arrow"],
    ) -> pa.Int32Array: ...
    def numbers(
        self,
        terms: Sequence[Hashable] | ArrayLike,
        missing: Literal["error", "negative", "null"] = "error",
        *,
        format: Literal["numpy", "arrow"] = "numpy",
    ) -> np.ndarray[tuple[int], np.dtype[np.int32]] | pa.Int32Array:
        "Look up the numbers for an array of terms or IDs."
        if pa.types.is_null(self._array.type):
            nums = pa.nulls(len(terms), type=pa.int32())
        else:
            term_arr = pa.array(terms, type=self._array.type)  # type: ignore
            nums = self._index.get_indexes(term_arr)  # type: ignore

        trace(self._log, "resolved %d IDs, %d invalid", len(terms), nums.null_count)
        if missing == "error" and nums.null_count:
            raise KeyError(f"{nums.null_count} invalid keys")
        elif missing == "negative":
            nums = pc.fill_null(nums, -1)  # type: ignore

        match format:
            case "numpy":
                return nums.to_numpy()
            case "arrow":
                return nums
            case _:  # pragma: nocover
                raise ValueError(f"invalid format {format}")

    def term(self, num: int) -> object:
        """
        Look up the term with a particular number.  Negative indexing is **not** supported.
        """
        if num < 0:
            raise IndexError("negative numbers not supported")
        return self._array[num].as_py()

    @overload
    def terms(
        self,
        nums: list[int] | NDArray[np.integer] | pd.Series | None = None,
        *,
        format: Literal["numpy"] = "numpy",
    ) -> IDArray: ...
    @overload
    def terms(
        self,
        nums: list[int] | NDArray[np.integer] | pd.Series | None = None,
        *,
        format: Literal["arrow"],
    ) -> pa.Array: ...
    def terms(
        self,
        nums: list[int] | NDArray[np.integer] | pd.Series | None = None,
        *,
        format: Literal["numpy", "arrow"] = "numpy",
    ) -> IDArray | pa.Array:
        """
        Get a list of terms, optionally for an array of term numbers.

        Args:
            nums:
                The numbers (indices) for of terms to retrieve.  If ``None``,
                returns all terms.

        Returns:
            The terms corresponding to the specified numbers, or the full array
            of terms (in order) if ``nums=None``.
        """
        ids = self._array
        if nums is not None:
            num_arr = pa.array(nums, type=pa.uint32())
            ids = ids.take(num_arr)

        match format:
            case "arrow":
                return ids
            case "numpy":
                return ids.to_numpy(zero_copy_only=False)
            case _:  # pragma: nocover
                raise ValueError(f"invalid format {format}")

    def id(self, num: int) -> object:
        "Alias for :meth:`term`  for greater readability for entity ID vocabularies."
        return self.term(num)

    @overload
    def ids(
        self,
        nums: list[int] | NDArray[np.integer] | pd.Series | None = None,
        *,
        format: Literal["numpy"] = "numpy",
    ) -> IDArray: ...
    @overload
    def ids(
        self,
        nums: list[int] | NDArray[np.integer] | pd.Series | None = None,
        *,
        format: Literal["arrow"],
    ) -> pa.Array: ...
    def ids(
        self,
        nums: list[int] | NDArray[np.integer] | pd.Series | None = None,
        *,
        format: Literal["numpy", "arrow"] = "numpy",
    ) -> IDArray | pa.Array:
        "Alias for :meth:`terms` for greater readability for entity ID vocabularies."
        return self.terms(nums, format=format)

    def id_array(self) -> pa.Array:
        return self._array

    def __eq__(self, other: Vocabulary) -> bool:  # noqa: F821
        if self is other:
            return True
        elif not isinstance(other, Vocabulary):
            return False
        else:
            self._ensure_hash()
            other._ensure_hash()
            return self._hash == other._hash

    def __contains__(self, key: object) -> bool:
        return self._index.get_index(key) is not None  # type: ignore

    def __iter__(self) -> Iterator[object]:
        return (t.as_py() for t in self._array)

    def __len__(self) -> int:
        return self.size

    def __array__(self, dtype=None) -> np.ndarray:
        return self._array.to_numpy(zero_copy_only=False)

    def __getstate__(self):
        return {"name": self.name, "array": self._array}

    def __setstate__(self, state):
        self.name = state["name"]
        self._array = state["array"]
        self._index = _accel.data.IDIndex(self._array)
        self._log = _log.bind(entity=self.name)

    def __str__(self) -> str:
        if self.name:
            return f"Vocabulary of {self.size} {self.name} terms"
        else:
            return f"Vocabulary of {self.size} terms"

    def __repr__(self) -> str:
        if self.name:
            return f"<Vocabulary(name={self.name}, size={self.size})>"
        else:
            return f"<Vocabulary(size={self.size})>"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import re
import tomllib
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from subprocess import CalledProcessError, check_output
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from packaging.version import Version

_log = logging.getLogger(__name__)
_lk_mod_dir = Path(__file__).parent
_lk_root = _lk_mod_dir.parent.parent


def is_git_install() -> bool:
    if _lk_mod_dir.parent.name != "src":
        return False

    git = _lk_mod_dir.parent.parent / ".git"
    return git.exists()


def lk_git_version() -> Version:
    from packaging.version import Version
    from packaging.version import parse as parse_version

    gd = check_output(["git", "describe", "--tags", "--match", "v*"], cwd=_lk_mod_dir.parent.parent)
    ver = gd.decode().strip()
    _log.debug("parsing Git info %s", ver)
    m = re.match(r"^v(\d+\.\d+\.\d+[.a-z0-9]*)(?:-(\d+)-(g[0-9a-fA-F]+))?$", ver)
    if not m:
        raise ValueError(f"unparseable version: {ver}")

    if m.group(2):
        pvs = f"{m.group(1)}.dev{m.group(2)}+{m.group(3)}"
    else:
        pvs = ver[1:]

    _log.debug("parsing transformed Git %s", pvs)
    version = parse_version(pvs)
    _log.info("Git version: %s", version)

    with open(_lk_root / "Cargo.toml", "rb") as tf:
        cargo = tomllib.load(tf)
    cv = cargo["package"]["version"]
    _log.debug("parsing Cargo version %s", cv)
    if m := re.match(r"(.*)-([a-z]+)(?:\.(\d+))(\.post\d+)", cv):
        cv_pr = m.group(2)
        if cv_pr != "rc":
            cv_pr = cv_pr[:1]
        cv_py = m.group(1) + m.group(2) + m.group(3) + m.group(4)
    else:
        cv_py = cv

    _log.debug("parsing transformed Cargo version %s", cv_py)
    cv_ver = parse_version(cv_py)
    _log.info("Cargo version: %s", cv_ver)

    if version.is_devrelease:
        if cv_ver > version:
            _log.debug("cargo requested version is newer")
            base = cv_ver.public
        else:
            _log.warning("Cargo version %s older than Git %s", cv_ver, version)
            if version.is_prerelease:
                assert version.pre is not None
                base = version.base_version + version.pre[0] + str(version.pre[1] + 1)
            else:
                base = f"{version.major}.{version.minor}.{version.micro}"

        version = Version(f"{base}.dev{version.dev}+{version.local}")
    else:
        if version != cv_ver:
            _log.warning("version mismatch: cargo %s != git %s", cv_ver, version)

    return version


def lenskit_version() -> str:
    if is_git_install():
        try:
            return str(lk_git_version())
        except CalledProcessError:
            pass

    try:
        return version("lenskit")
    except PackageNotFoundError:  # pragma: nocover
        return "UNKNOWN"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import re
from dataclasses import dataclass
from types import NoneType

import numpy as np
from typing_extensions import assert_type

from pytest import fail, mark, warns

from lenskit.pipeline import PipelineBuilder, PipelineWarning
from lenskit.pipeline.components import Component
from lenskit.pipeline.config import PipelineConfig
from lenskit.pipeline.nodes import (
    ComponentConstructorNode,
    ComponentInstanceNode,
    ComponentNode,
    InputNode,
)

_log = logging.getLogger(__name__)


# region Test Components
@dataclass
class PrefixConfig:
    prefix: str


class Prefixer(Component[str]):
    config: PrefixConfig

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


def negative(x: int) -> int:
    return -x


def double(x: int) -> int:
    return x * 2


def add(x: int | np.ndarray, y: int) -> int | np.ndarray:
    return x + y


def msg_ident(msg: str) -> str:
    return msg


def msg_prefix(prefix: str, msg: str) -> str:
    return prefix + msg


# endregion


def test_serialize_input():
    "serialize with one input node"
    pipe = PipelineBuilder("test")
    pipe.create_input("user", int, str)

    cfg = pipe.build_config()
    print(cfg)
    assert cfg.meta.name == "test"
    assert len(cfg.inputs) == 1
    assert cfg.inputs[0].name == "user"
    assert cfg.inputs[0].types == {"int", "str"}


def test_round_trip_input():
    "serialize with one input node"
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str)

    cfg = pipe.build_config()
    print(cfg)

    p2 = PipelineBuilder.from_config(cfg)
    i2 = p2.node("user")
    assert isinstance(i2, InputNode)
    assert i2.name == "user"
    assert i2.types == {int, str}


def test_round_trip_optional_input():
    "serialize with one input node"
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str, None)

    cfg = pipe.build_config()
    assert cfg.inputs[0].types == {"int", "str", "None"}

    p2 = PipelineBuilder.from_config(cfg)
    i2 = p2.node("user")
    assert isinstance(i2, InputNode)
    assert i2.name == "user"
    assert i2.types == {int, str, NoneType}


def test_config_single_node():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    pipe.add_component("return", msg_ident, msg=msg)

    cfg = pipe.build_config()
    assert len(cfg.inputs) == 1
    assert len(cfg.components) == 1

    assert cfg.components["return"].code.endswith("test_save_load:msg_ident")
    assert cfg.components["return"].config is None
    assert cfg.components["return"].inputs == {"msg": "msg"}


def test_round_trip_single_node():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    pipe.add_component("return", msg_ident, msg=msg)

    cfg = pipe.build_config()

    p2 = PipelineBuilder.from_config(cfg)
    assert len(p2.nodes()) == 2
    r2 = p2.node("return")
    assert isinstance(r2, ComponentInstanceNode)
    assert r2.component is msg_ident
    assert p2._edges["return"] == {"msg": "msg"}

    p2 = p2.build()
    assert p2.run("return", msg="foo") == "foo"


def test_configurable_component():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    pfx = Prefixer(prefix="scroll named ")
    pipe.add_component("prefix", pfx, msg=msg)

    cfg = pipe.build_config()
    assert cfg.components["prefix"].config == {"prefix": "scroll named "}

    p2 = PipelineBuilder.from_config(cfg).build()
    assert len(p2.nodes()) == 2
    r2 = p2.node("prefix")
    assert isinstance(r2, ComponentInstanceNode)
    assert isinstance(r2.component, Prefixer)
    assert r2.component is not pfx
    assert p2._edges["prefix"] == {"msg": "msg"}

    assert p2.run("prefix", msg="HACKEM MUCHE") == "scroll named HACKEM MUCHE"

    print("hash:", pipe.config_hash)
    assert pipe.config_hash is not None
    assert p2.config_hash == pipe.config_hash()


def test_save_with_defaults():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)
    pipe.default_connection("msg", msg)
    pipe.add_component("return", msg_ident)
    pipe.default_component("return")

    cfg = pipe.build_config()

    pipe = pipe.build()
    assert pipe.run(msg="hello") == "hello"

    p2 = PipelineBuilder.from_config(cfg)

    p2 = p2.build()
    assert p2.run(msg="hello") == "hello"


def test_hashes_different():
    p1 = PipelineBuilder()
    p2 = PipelineBuilder()

    a1 = p1.create_input("a", int)
    a2 = p2.create_input("a", int)

    # at this point the hashes should be the same
    _log.info("p1 stage 1 hash: %s", p1.config_hash())
    _log.info("p2 stage 1 hash: %s", p2.config_hash())
    assert p1.config_hash() == p2.config_hash()

    p1.add_component("proc", negative, x=a1)
    p2.add_component("proc", double, x=a2)

    # with different components, they should be different
    _log.info("p1 stage 2 hash: %s", p1.config_hash())
    _log.info("p2 stage 2 hash: %s", p2.config_hash())
    assert p1.config_hash() != p2.config_hash()
    assert p1.build().config_hash != p2.build().config_hash


def test_save_with_fallback():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    nd = pipe.add_component("double", double, x=a)
    nn = pipe.add_component("negate", negative, x=a)
    fb = pipe.use_first_of("fill-operand", b, nn)
    pipe.add_component("add", add, x=nd, y=fb)

    cfg = pipe.build_config()
    json = cfg.model_dump_json(exclude_none=True)
    print(json)
    c2 = PipelineConfig.model_validate_json(json)

    p2 = PipelineBuilder.from_config(c2)

    p2 = p2.build()
    # 3 * 2 + -3 = 3
    assert p2.run(("fill-operand", "add"), a=3) == (-3, 3)


@mark.skip("hash checks disabled")
def test_hash_validate():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    pfx = Prefixer(prefix="scroll named ")
    pipe.add_component("prefix", pfx, msg=msg)

    cfg = pipe.build_config()
    print("initial config:", cfg.model_dump_json(indent=2))
    assert cfg.meta.hash is not None
    cfg.components["prefix"].config["prefix"] = "scroll called "  # type: ignore
    print("modified config:", cfg.model_dump_json(indent=2))

    with warns(PipelineWarning):
        PipelineBuilder.from_config(cfg)


def test_alias_input():
    "just an input node and an alias"
    pipe = PipelineBuilder()
    user = pipe.create_input("user", int, str)

    pipe.alias("person", user)

    cfg = pipe.build_config()

    p2 = PipelineBuilder.from_config(cfg)
    p2 = p2.build()
    assert p2.run("person", user=32) == 32


def test_alias_node():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)
    pipe.alias("result", na)

    pipe = pipe.build()
    assert pipe.run("result", a=5, b=7) == 17

    p2 = pipe.clone()
    assert p2.run("result", a=5, b=7) == 17


def test_literal():
    pipe = PipelineBuilder("literal-prefix")
    msg = pipe.create_input("msg", str)

    pipe.add_component("prefix", msg_prefix, prefix=pipe.literal("hello, "), msg=msg)
    pipe.default_component("prefix")

    pipe = pipe.build()
    assert pipe.run(msg="HACKEM MUCHE") == "hello, HACKEM MUCHE"

    print(pipe.config.model_dump_json(indent=2))
    p2 = pipe.clone()
    assert p2.run(msg="FOOBIE BLETCH") == "hello, FOOBIE BLETCH"


def test_literal_array():
    pipe = PipelineBuilder("literal-add-array")
    a = pipe.create_input("a", int)

    pipe.add_component("add", add, x=np.arange(10), y=a)
    pipe.default_component("add")

    print("pipeline:", pipe.build_config().model_dump_json(indent=2))
    pipe = pipe.build()
    res = pipe.run(a=5)
    assert np.all(res == np.arange(5, 15))

    p2 = pipe.clone()
    assert np.all(p2.run(a=5) == np.arange(5, 15))


def test_stable_with_literals():
    "test that two identical pipelines have the same hash, even with literals"
    p1 = PipelineBuilder("literal-add-array")
    a = p1.create_input("a", int)
    p1.add_component("add", add, x=np.arange(10), y=a)

    p2 = PipelineBuilder("literal-add-array")
    a = p2.create_input("a", int)
    p2.add_component("add", add, x=np.arange(10), y=a)

    assert p1.config_hash() == p2.config_hash()
    assert p1.build().config_hash == p2.build().config_hash



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np

from pytest import importorskip, mark

from lenskit.data import ItemList, from_interactions_df
from lenskit.pipeline import topn_pipeline
from lenskit.testing import BasicComponentTests, ScorerTests
from lenskit.training import TrainingOptions

hpf = importorskip("lenskit.hpf")

_log = logging.getLogger(__name__)


class TestHPF(BasicComponentTests, ScorerTests):
    component = hpf.HPFScorer


@mark.slow
def test_hpf_train_large(tmp_path, ml_ratings):
    algo = hpf.HPFScorer(features=20)
    ratings = ml_ratings.assign(rating=ml_ratings.rating + 0.5)
    ds = from_interactions_df(ratings)
    algo.train(ds)

    assert algo.user_features_.shape[0] == ratings.user_id.nunique()
    assert algo.item_features_.shape[0] == ratings.item_id.nunique()

    mfile = tmp_path / "hpf.dat"
    with mfile.open("wb") as mf:
        pickle.dump(algo, mf)

    with mfile.open("rb") as mf:
        a2 = pickle.load(mf)

    assert np.all(a2.user_features_ == algo.user_features_)
    assert np.all(a2.item_features_ == algo.item_features_)

    pipe = topn_pipeline(algo)
    pipe.train(ds, TrainingOptions(retrain=False))

    for u in np.random.choice(ratings.user_id.unique(), size=50, replace=False):
        recs = pipe.run("recommender", query=u, n=50)
        assert isinstance(recs, ItemList)
        assert len(recs) == 50


@mark.slow
def test_hpf_train_binary(tmp_path, ml_ratings):
    algo = hpf.HPFScorer(features=20)
    ratings = ml_ratings.drop(columns=["timestamp", "rating"])
    ds = from_interactions_df(ratings)
    algo.train(ds)

    assert algo.user_features_.shape[0] == ratings.user_id.nunique()
    assert algo.item_features_.shape[0] == ratings.item_id.nunique()

    mfile = tmp_path / "hpf.dat"
    with mfile.open("wb") as mf:
        pickle.dump(algo, mf)

    with mfile.open("rb") as mf:
        a2 = pickle.load(mf)

    assert np.all(a2.user_features_ == algo.user_features_)
    assert np.all(a2.item_features_ == algo.item_features_)

    pipe = topn_pipeline(algo)
    pipe.train(ds, TrainingOptions(retrain=False))

    for u in np.random.choice(ratings.user_id.unique(), size=50, replace=False):
        recs = pipe.run("recommender", query=u, n=50)
        assert isinstance(recs, ItemList)
        assert len(recs) == 50



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path
from subprocess import check_call

import click
from humanize import metric

from lenskit.logging import Task, get_logger, stdout_console

_log = get_logger(__name__)
_gh_out: Path | None = None


@click.command("task")
@click.option(
    "-l",
    "--label",
    metavar="LABEL",
    default="cli-task",
    help="Human-readable task label",
)
@click.argument("args", nargs=-1)
def task(label: str, args: list[str]):  # pragma: nocover
    """
    Run a task with LensKit task tracking.
    """
    console = stdout_console()

    log = _log.bind(label=label)

    with Task(label) as task:
        log.info("running command %s", " ".join(args))
        check_call(args)

    if task.system_power is not None:
        log.info("task consumed %s", metric(task.system_power / 3600, "Wh"))
    console.print(task)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Support for parallelism with Ray.
"""

from __future__ import annotations

import base64
import itertools
import os
import pickle
from collections import deque
from collections.abc import Callable, Generator, Iterable, Iterator, Sequence
from functools import partial
from platform import python_version_tuple
from typing import Any, Generic, TypeVar

from lenskit.logging import Task, get_logger
from lenskit.logging.worker import WorkerContext, WorkerLogConfig

from .config import (
    ParallelConfig,
    effective_cpu_count,
    ensure_parallel_init,
    get_parallel_config,
    subprocess_config,
)
from .invoker import A, InvokeOp, M, ModelOpInvoker, R

try:
    import ray
    from ray.remote_function import RemoteFunction

    _ray_imported = True
except ImportError:
    _ray_imported = False


T = TypeVar("T")

BATCH_SIZE = 200
_worker_parallel: ParallelConfig
_log = get_logger(__name__)


def ensure_cluster():
    import ray

    if not ray.is_initialized():
        _log.debug("Ray is not initialized, initializing")
        try:
            init_cluster()
        except ValueError as e:
            _log.debug("Ray initialization failed", exception=e)
            if "existing cluster" in str(e):
                _log.info("Ray cluster already started, reusing")
                _log.warn("existing Ray cluster may not properly throttle LensKit jobs")
                ray.init()
            else:
                raise e


def init_cluster(
    *,
    num_cpus: int | None = None,
    proc_slots: int | None = None,
    resources: dict[str, float] | None = None,
    worker_parallel: ParallelConfig | None = None,
    global_logging: bool = False,
    **kwargs,
):
    """
    Initialize or connect to a Ray cluster, with the LensKit options.

    The resulting cluster can be used by an invoker, or it can be used directly.
    The Ray invoker uses batching, though, so it only works well with many small
    tasks.

    Args:
        num_cpus:
            The total number of CPUs to allow. Defaults to
            :fun:`effective_cpu_count`.
        proc_slots:
            The number of “process slots” for LensKit parallel operations.
            Defaults to the LensKit process count.  These slots are recorded as
            the ``lk_process`` resource on the Ray cluster.
        resources:
            Additional custom resources to register in the Ray cluster.
        worker_parallel:
            Parallel processing configuration for worker processes.  If
            ``None``, uses the default.
        global_logging:
            ``True`` to wire up logging in the workers at startup, instead of only
            connecting logs when a task is run.
        kwargs:
            Other options to pass to :func:`ray.init`.

    Stability:
        Experimental
    """
    global _worker_parallel
    import ray
    import ray.runtime_env

    if resources is None:
        resources = {}
    else:
        resources = resources.copy()

    cfg = get_parallel_config()
    if proc_slots is None:
        proc_slots = cfg.processes

    if num_cpus is None:
        num_cpus = effective_cpu_count()

    if worker_parallel is None:
        worker_parallel = subprocess_config()
    _worker_parallel = worker_parallel

    env = worker_parallel.env_vars().copy()
    wc = WorkerLogConfig.current()
    env["LK_LOG_CONFIG"] = base64.encodebytes(pickle.dumps(wc)).decode()

    setup = _worker_setup if global_logging else None
    runtime = ray.runtime_env.RuntimeEnv(env_vars=env, worker_process_setup_hook=setup)

    _log.info("starting Ray cluster")
    ray.init(num_cpus=num_cpus, resources=resources, runtime_env=runtime, **kwargs)


def inference_worker_cpus() -> int:
    return _worker_parallel.backend_threads


def training_worker_cpus() -> int:
    return _worker_parallel.total_threads


def ray_supported() -> bool:
    """
    Check if this Ray setup is supported by LensKit.
    """
    major, minor, patch = python_version_tuple()
    if int(major) < 3 or int(minor) < 12:
        return False
    else:
        return ray_available()


def ray_available() -> bool:
    """
    Check if Ray is available.
    """
    try:
        import ray  # noqa: F401

        return True
    except ImportError:
        return False


def ray_active() -> bool:
    """
    Query whether Ray is active.
    """
    if ray_available():
        import ray

        return ray.is_initialized()
    else:
        return False


def is_ray_worker() -> bool:
    """
    Determine whether the current process is running on a Ray worker.
    """
    # logic adapted from https://discuss.ray.io/t/how-to-know-if-code-is-running-on-ray-worker/15642
    if ray_active():
        ctx = ray.get_runtime_context()
        return ctx.worker.mode == ray.WORKER_MODE
    else:
        return False


class RayOpInvoker(ModelOpInvoker[A, R], Generic[M, A, R]):
    function: InvokeOp[M, A, R]
    model_ref: Any
    action: RemoteFunction
    limit: int | None

    def __init__(self, model: M, func: InvokeOp[M, A, R], *, limit: int | None = None):
        _log.debug("persisting to Ray cluster")
        import ray
        import torch

        if isinstance(model, ray.ObjectRef):
            self.model_ref = model
        else:
            self.model_ref = ray.put(model)
        self.function = func

        self.limit = limit

        worker = ray.remote(_ray_invoke_worker)
        # request a little GPU
        if torch.cuda.is_available():
            n_gpus = 0.1
        else:
            n_gpus = None
        self.action = worker.options(num_cpus=inference_worker_cpus(), num_gpus=n_gpus)  # type: ignore

    def map(self, tasks: Iterable[A]) -> Iterator[R]:
        limit = TaskLimiter(self.limit)
        function = partial(self.action.remote, self.function, self.model_ref)
        for bres in limit.imap(function, itertools.batched(tasks, BATCH_SIZE)):
            yield from bres  # type: ignore

    def shutdown(self):
        del self.model_ref


class TaskLimiter:
    """
    Limit task concurrency using :func:`ray.wait`.

    This class provides two key operations:

    - Add a task to the limiter with :meth:`add_task`.
    - Wait for tasks until the number of pending tasks is less than the limit.
    - Wait for all tasks with `drain`.

    Args:
        limit:
            The maximum number of pending tasks.  Defaults to the LensKit
            process count (see :func:`lenskit.parallel.initialize`).
    """

    limit: int
    "The maximum number of pending tasks."
    finished: int
    "The number of tasks completed."

    _tasks: list[Any]

    def __init__(self, limit: int | None = None):
        if limit is None or limit <= 0:
            self.limit = get_parallel_config().processes
        else:
            self.limit = limit
        self.finished = 0
        self._tasks = []

    @property
    def pending(self) -> int:
        """
        The number of pending tasks.
        """
        return len(self._tasks)

    def imap(
        self,
        function: RemoteFunction | Callable[[A], R],
        items: Iterable[A],
        *,
        ordered: bool = True,
    ) -> Generator[R, None, int]:
        if ordered:
            return self._imap_ordered(function, items)
        else:
            return self._imap_unordered(function, items)

    def _imap_ordered(
        self, function: RemoteFunction | Callable[[A], R], items: Iterable[A]
    ) -> Generator[Any, None, int]:
        n = 0
        queued = deque()
        ready = set()

        # make a callable from a Ray remote function
        if hasattr(function, "remote"):
            function = function.remote  # type: ignore

        for arg in items:
            for res in self.results_until_limit():
                ready.add(res)
                n += 1

            yield from _drain_queue(queued, ready)

            task: Any = function(arg)
            self.add_task(task)
            queued.append(task)

        for res in self.drain_results():
            ready.add(res)
            n += 1

            yield from _drain_queue(queued, ready)

        assert len(queued) == 0
        assert len(ready) == 0

        return n

    def _imap_unordered(
        self, function: RemoteFunction | Callable[[A], R], items: Iterable[A]
    ) -> Generator[Any, None, int]:
        n = 0

        # make a callable from a Ray remote function
        if hasattr(function, "remote"):
            function = function.remote  # type: ignore

        for arg in items:
            for res in self.results_until_limit():
                yield ray.get(res)
                n += 1
            task: Any = function(arg)
            self.add_task(task)

        for res in self.drain_results():
            yield ray.get(res)
            n += 1

        return n

    def add_task(self, task: ray.ObjectRef | ray.ObjectID):
        self._tasks.append(task)

    def results_until_limit(self) -> Generator[ray.ObjectRef, None, None]:
        """
        Iterate over available results until the number of pending results tasks
        is less than the limit, blocking as needed.

        This is a generator, returning the task result references.  The iterator
        will stop when the pending tasks list is under the limit.  No guarantee
        is made on the order of returned results.
        """
        return self._drain_until(self.limit - 1)

    def wait_for_limit(self):
        """
        Wait until the pending tasks are back under the limit.

        This method calls :meth:`ray.get` on the result of each pending task to
        resolve errors, but discards the return value.
        """
        for r in self.results_until_limit():
            ray.get(r)

    def drain_results(self) -> Generator[ray.ObjectRef, None, None]:
        """
        Iterate over all remaining tasks until the pending task list is empty,
        blocking as needed.

        This is a generator, returning the task result references.  No guarantee
        is made on the order of returned results.
        """
        return self._drain_until(0)

    def drain(self):
        """
        Wait until all pending tasks are finished.

        This method calls :meth:`ray.get` on the result of each pending task to
        resolve errors, but discards the return value.
        """
        for r in self.drain_results():
            ray.get(r)

    def _drain_until(self, limit: int) -> Generator[ray.ObjectRef, None, None]:
        while self._tasks and len(self._tasks) > limit:
            done, remaining = ray.wait(self._tasks)
            self._tasks = remaining
            for dt in done:
                self.finished += 1
                yield dt


def _drain_queue(queued: deque[ray.ObjectRef], ready: set[ray.ObjectRef]) -> Iterator[Any]:
    while queued and queued[0] in ready:
        ref = queued.popleft()
        ready.remove(ref)
        yield ray.get(ref)


def _ray_invoke_worker(func: Callable[[M, A], R], model: M, args: Sequence[A]) -> list[R]:
    with init_worker(autostart=False) as ctx:
        try:
            with Task("cluster worker", subprocess=True) as task:
                result = [func(model, arg) for arg in args]
        finally:
            ctx.send_task(task)

    return result


def _worker_setup():
    init_worker()


def init_worker(*, autostart: bool = True) -> WorkerContext:
    """
    Initialize a Ray worker process.  Sets up logging, and returns the context.

    Args:
        autostart:
            Set to ``False`` to disable calling :meth:`WorkerContext.start`, for
            when the caller will start and stop the context if it is new.
    """
    log_cfg = pickle.loads(base64.decodebytes(os.environb[b"LK_LOG_CONFIG"]))
    context = WorkerContext.active()
    if context is None:
        context = WorkerContext(log_cfg)
        if autostart:
            context.start()

    ensure_parallel_init()
    return context



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
from scipy.sparse import csr_array

from pytest import importorskip, mark

from lenskit.data import Dataset
from lenskit.testing import ml_20m

numba = importorskip("numba")


@mark.benchmark()
def test_select_arrow(ml_20m: Dataset, rng: np.random.Generator, benchmark):
    matrix = ml_20m.interactions().matrix().scipy(layout="csr")

    offsets = np.require(matrix.indptr[:-1], np.int32)
    offsets = pa.array(offsets)
    pairs = pa.StructArray.from_arrays(
        [pa.array(matrix.indices), pa.array(matrix.data)], ["column", "value"]
    )
    lists = pa.ListArray.from_arrays(offsets, pairs)

    users = rng.choice(ml_20m.user_count, 1000, replace=True)
    users = np.require(users, "i4")
    users = pa.array(users)

    def select():
        _res = pc.array_take(lists, users, boundscheck=False)

    benchmark(select)


@mark.benchmark()
def test_select_scipy_csr(ml_20m: Dataset, rng: np.random.Generator, benchmark):
    matrix = ml_20m.interactions().matrix().scipy(layout="csr")

    users = rng.choice(ml_20m.user_count, 1000, replace=True)
    users = np.require(users, "i4")

    def select():
        _res = matrix[users, :]

    benchmark(select)


@mark.benchmark()
def test_select_numba(ml_20m: Dataset, rng: np.random.Generator, benchmark):
    matrix = ml_20m.interactions().matrix().scipy(layout="csr")

    users = rng.choice(ml_20m.user_count, 1000, replace=True)
    users = np.require(users, "i4")
    numba_take(matrix, users)

    def select():
        _res = numba_take(matrix, users)

    benchmark(select)


def numba_take(matrix: csr_array, rows: np.ndarray[tuple[int], np.dtype[np.int32]]):
    assert matrix.shape is not None

    _nr, nc = matrix.shape
    rp2, ci2, vs2 = _select_matrix(matrix.indptr, matrix.indices, matrix.data, rows)
    return csr_array((vs2, ci2, rp2), shape=(len(rows), nc))


@numba.njit
def _select_matrix(
    rps: np.ndarray[tuple[int], np.dtype[np.int32]],
    cis: np.ndarray[tuple[int], np.dtype[np.int32]],
    vs: np.ndarray[tuple[int], np.dtype[np.float32]],
    rows: np.ndarray[tuple[int], np.dtype[np.int32]],
) -> tuple[
    np.ndarray[tuple[int], np.dtype[np.int32]],
    np.ndarray[tuple[int], np.dtype[np.int32]],
    np.ndarray[tuple[int], np.dtype[np.float32]],
]:
    ptrs = np.zeros(len(rows) + 1, np.int32)
    tot = 0
    for i, r in enumerate(rows):
        sp = rps[r]
        ep = rps[r + 1]
        rl = ep - sp
        ptrs[i + 1] = ptrs[i] + rl
        tot += rl

    inds = np.empty(tot, np.int32)
    vals = np.empty(tot, np.float32)

    for i, r in enumerate(rows):
        sp = rps[r]
        ep = rps[r + 1]

        spo = ptrs[i]
        epo = ptrs[i + 1]
        inds[spo:epo] = cis[sp:ep]
        vals[spo:epo] = vs[sp:ep]

    return ptrs, inds, vals



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Test extremely large datasets.
"""

import os
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import mark

from lenskit.data import DatasetBuilder
from lenskit.data.adapt import from_interactions_df
from lenskit.data.movielens import load_movielens_df
from lenskit.logging import get_logger

pytestmark = mark.skipif("LK_HUGE_TEST" not in os.environ, reason="huge tests disabled")

_log = get_logger(__name__)
huge_dir = Path("data/ml-20mx16x32")
ml32m_zip = Path("data/ml-32m.zip")


@mark.skipif(not ml32m_zip.exists(), reason="ML 32M not available")
@mark.slow
def test_basic_320M():
    "Test building a 32M data set"

    _log.info("loading ML32M dataframe")
    df = load_movielens_df(ml32m_zip)
    _log.info("creating ML32M dataset")
    ds32 = from_interactions_df(df)

    dsb = DatasetBuilder()
    dsb.add_relationship_class(
        "interaction", ["user", "item"], allow_repeats=False, interaction=True
    )

    for i in range(10):
        _log.info("adding batch %d", i)
        df_part = pd.DataFrame(
            {
                "user_id": df["user_id"] + (i * 500000),
                "item_id": df["item_id"],
            }
        )
        dsb.add_interactions("interaction", df_part, missing="insert")

    _log.info("building dataset")
    ds = dsb.build()
    assert ds.interaction_count == len(df) * 10
    assert ds.user_count == ds32.user_count * 10
    assert ds.item_count == ds32.item_count

    istats32 = ds32.item_stats()
    istats = ds.item_stats()
    assert np.all(istats["count"] == istats32["count"] * 10)
    ustats = ds.user_stats()
    assert len(ustats) == ds.user_count


@mark.skip("1B does not work")
@mark.skipif(not huge_dir.exists(), reason="ML 1B not available")
@mark.slow
def test_basic_1B():
    "Test building a 1B data set"

    dsb = DatasetBuilder()
    dsb.add_relationship_class(
        "interaction", ["user", "item"], allow_repeats=False, interaction=True
    )

    n_users = 0
    n_rows = 0
    batches = []
    for file in huge_dir.glob("train*.npz"):
        _log.info("loading segment", file=file.name)
        npz = np.load(file)
        array = next(iter(npz.values()))
        n_users += len(np.unique(array[:, 0]))
        n_rows += array.shape[0]
        batches.append(
            pa.record_batch(
                [pa.array(array[:, 0], pa.int32()), pa.array(array[:, 1], pa.int32())],
                names=["user_id", "item_id"],
            )
        )
        del array
        # dsb.add_interactions(
        #     "interaction",
        #     pd.DataFrame(array, columns=["user_id", "item_id"]),
        #     missing="insert",
        # )

    _log.info("making table")
    table = pa.Table.from_batches(batches)
    del batches
    _log.info("adding to dataset")
    dsb.add_interactions("interaction", table, missing="insert")

    _log.info("building dataset")
    ds = dsb.build()
    assert ds.interaction_count == n_rows
    assert ds.user_count == n_users



"""
Rust acceleration code.
"""

from __future__ import annotations

import pyarrow as pa

from lenskit.data.types import NPMatrix, NPVector
from lenskit.funksvd import FunkSVDTrainingData, FunkSVDTrainingParams

from . import als, data

__all__ = [
    "als",
    "data",
    "init_accel_pool",
    "thread_count",
    "FunkSVDTrainer",
]

def init_accel_pool(n_threads: int): ...
def thread_count() -> int: ...

class FunkSVDTrainer:
    def __init__(
        self,
        config: FunkSVDTrainingParams,
        data: FunkSVDTrainingData,
        user_features: NPMatrix,
        item_features: NPMatrix,
    ): ...
    def feature_epoch(self, feature: int, estimates: NPVector, trail: float) -> float: ...

def sparse_row_debug_type(arr: pa.Array) -> tuple[str, int, int]: ...
def sparse_structure_debug_large(arr: pa.Array) -> tuple[int, int, int]: ...
